"""Gumbel module."""

import numpy as np

from sdgx.models.components.sdv_copulas.bivariate.base import Bivariate, CopulaTypes
from sdgx.models.components.sdv_copulas.bivariate.utils import split_matrix


class Gumbel(Bivariate):
    """Class for clayton copula model."""

    copula_type = CopulaTypes.GUMBEL
    theta_interval = [1, float("inf")]
    invalid_thetas = []

    def generator(self, t):
        """Return the generator function."""
        return np.power(-np.log(t), self.theta)

    def probability_density(self, X):
        r"""Compute probability density function for given copula family.

        The probability density(PDF) for the Gumbel family of copulas correspond to the formula:

        .. math::

            \begin{align}
                c(U,V)
                    &= \frac{\partial^2 C(u,v)}{\partial v \partial u}
                    &= \frac{C(u,v)}{uv} \frac{((-\ln u)^{\theta}  # noqa: JS101
                    + (-\ln v)^{\theta})^{\frac{2}  # noqa: JS101
                {\theta} - 2 }}{(\ln u \ln v)^{1 - \theta}}  # noqa: JS101
                ( 1 + (\theta-1) \big((-\ln u)^\theta
                + (-\ln v)^\theta\big)^{-1/\theta})
            \end{align}

        Args:
            X (numpy.ndarray)

        Returns:
            numpy.ndarray

        """
        self.check_fit()

        U, V = split_matrix(X)

        if self.theta == 1:
            return U * V

        else:
            a = np.power(U * V, -1)
            tmp = np.power(-np.log(U), self.theta) + np.power(-np.log(V), self.theta)
            b = np.power(tmp, -2 + 2.0 / self.theta)
            c = np.power(np.log(U) * np.log(V), self.theta - 1)
            d = 1 + (self.theta - 1) * np.power(tmp, -1.0 / self.theta)
            return self.cumulative_distribution(X) * a * b * c * d

    def cumulative_distribution(self, X):
        r"""Compute the cumulative distribution function for the Gumbel copula.

        The cumulative density(cdf), or distribution function for the Gumbel family of copulas
        correspond to the formula:

        .. math:: C(u,v) = e^{-((-\ln u)^{\theta} + (-\ln v)^{\theta})^{\frac{1}{\theta}}}

        Args:
            X (np.ndarray)

        Returns:
            np.ndarray: cumulative probability for the given datapoints, cdf(X).

        """
        self.check_fit()

        U, V = split_matrix(X)

        if self.theta == 1:
            return U * V

        else:
            h = np.power(-np.log(U), self.theta) + np.power(-np.log(V), self.theta)
            h = -np.power(h, 1.0 / self.theta)
            cdfs = np.exp(h)
            return cdfs

    def percent_point(self, y, V):
        """Compute the inverse of conditional cumulative distribution :math:`C(u|v)^{-1}`.

        Args:
            y (np.ndarray): value of :math:`C(u|v)`.
            v (np.ndarray): given value of v.

        """
        self.check_fit()

        if self.theta == 1:
            return y

        else:
            return super().percent_point(y, V)

    def partial_derivative(self, X):
        r"""Compute partial derivative of cumulative distribution.

        The partial derivative of the copula(CDF) is the conditional CDF.

        .. math:: F(v|u) = \frac{\partial C(u,v)}{\partial u} =
            C(u,v)\frac{((-\ln u)^{\theta} + (-\ln v)^{\theta})^{\frac{1}{\theta} - 1}}
            {\theta(- \ln u)^{1 -\theta}}

        Args:
            X (np.ndarray)
            y (float)

        Returns:
            numpy.ndarray

        """
        self.check_fit()

        U, V = split_matrix(X)

        if self.theta == 1:
            return V

        else:
            t1 = np.power(-np.log(U), self.theta)
            t2 = np.power(-np.log(V), self.theta)
            p1 = self.cumulative_distribution(X)
            p2 = np.power(t1 + t2, -1 + 1.0 / self.theta)
            p3 = np.power(-np.log(V), self.theta - 1)
            return p1 * p2 * p3 / V

    def compute_theta(self):
        r"""Compute theta parameter using Kendall's tau.

        On Gumbel copula :math:`\tau` is defined as :math:`τ = \frac{θ−1}{θ}`
        that we solve as :math:`θ = \frac{1}{1-τ}`
        """
        if self.tau == 1:
            raise ValueError("Tau value can't be 1")

        return 1 / (1 - self.tau)



import pandas as pd
import pytest

from sdgx.data_models.inspectors.subset_relationship import SubsetRelationshipInspector
from sdgx.data_models.metadata import Metadata
from sdgx.data_models.relationship import Relationship


@pytest.fixture
def inspector():
    yield SubsetRelationshipInspector()


@pytest.fixture
def dummy_data(demo_relational_table_path):
    table_path_a, table_path_b, _ = demo_relational_table_path
    df_a = pd.read_csv(table_path_a)
    df_b = pd.read_csv(table_path_b)

    yield [
        (df_a, "parent", Metadata.from_dataframe(df_a)),
        (df_b, "child", Metadata.from_dataframe(df_b)),
    ]


@pytest.fixture
def dummy_relationship(demo_relational_table_path):
    _, _, pairs = demo_relational_table_path

    yield Relationship.build(
        parent_table="parent",
        child_table="child",
        foreign_keys=pairs,
    )


def test_inspector(dummy_data, dummy_relationship, inspector: SubsetRelationshipInspector):
    for raw_data, name, metadata in dummy_data:
        inspector.fit(raw_data, name=name, metadata=metadata)
    relationships = inspector.inspect()["relationships"]
    assert relationships
    assert relationships == [dummy_relationship]


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import numpy as np
import pandas as pd
from scipy.stats import entropy
from sklearn.metrics.cluster import normalized_mutual_info_score

from sdgx.metrics.multi_table.base import MultiTableMetric
from sdgx.metrics.pair_column.mi_sim import MISim


class MISim(MultiTableMetric):
    """MISim : Mutual Information Similarity

    This class is used to calculate the Mutual Information Similarity between the target columns of real data and synthetic data.

    Currently, we support discrete and continuous(need to be discretized) columns as inputs.
    """

    def __init__(self) -> None:
        super().__init__()
        self.lower_bound = 0
        self.upper_bound = 1
        self.metric_name = "mutual_information_similarity"
        self.numerical_bins = 50

    @classmethod
    def calculate(
        real_data: pd.DataFrame, synthetic_data: pd.DataFrame, metadata: dict
    ) -> pd.DataFrame:
        """
        Calculate the Mutual Information Similarity between a real column and a synthetic column.
        Args:
            real_data (pd.DataFrame): The real data.
            synthetic_data (pd.DataFrame): The synthetic data.
            metadata(dict): The metadata that describes the data type of each column

        Returns:
            MI_similarity (float): The metric value.
        """

        # 传入概率分布数组

        columns = synthetic_data.columns
        n = len(columns)
        mi_sim_instance = MISim()
        nMI_sim = np.zeros((n, n))

        for i in range(len(columns)):
            for j in range(len(columns)):
                syn_data = pd.concat(
                    [synthetic_data[columns[i]], synthetic_data[columns[j]]], axis=1
                )
                real_data = pd.concat([real_data[columns[i]], real_data[columns[j]]], axis=1)

                nMI_sim[i][j] = mi_sim_instance.calculate(real_data, syn_data, metadata)

        MI_sim = np.sum(nMI_sim) / n / n
        # test
        MISim.check_output(MI_sim)

        return MI_sim

    @classmethod
    def check_output(cls, raw_metric_value: float):
        """Check the output value.

        Args:
            raw_metric_value (float):  the calculated raw value of the Mutual Information Similarity.
        """
        instance = cls()
        if raw_metric_value < instance.lower_bound or raw_metric_value > instance.upper_bound:
            raise ValueError



from __future__ import annotations

import shutil
from pathlib import Path
from typing import Generator, List

import pandas as pd

from sdgx.cachers.base import Cacher
from sdgx.cachers.extension import hookimpl
from sdgx.data_connectors.base import DataConnector
from sdgx.exceptions import CacheError
from sdgx.utils import logger


class DiskCache(Cacher):
    """
    Cacher that cache data in disk with parquet format

    Args:
        blocksize (int): The blocksize of the cache.
        cache_dir (str | Path | None, optional): The directory where the cache will be stored. Defaults to None.
        identity (str | None, optional): The identity of the data source. Defaults to None.

    Todo:
        * Add partial cache when blocksize > chunksize
        * Improve cache invalidation
        * Improve performance if blocksize > chunksize
    """

    def __init__(
        self,
        cache_dir: str | Path | None = None,
        identity: str | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        if not cache_dir:
            cache_dir = Path.cwd() / ".sdgx_cache"
            if identity:
                cache_dir = cache_dir / identity
        self.cache_dir = Path(cache_dir)
        self.cache_dir.mkdir(parents=True, exist_ok=True)

    def clear_cache(self) -> None:
        """
        Clear all cache in cache_dir.
        """
        for f in self.cache_dir.glob("*.parquet"):
            f.unlink()
        shutil.rmtree(self.cache_dir, ignore_errors=True)

    def clear_invalid_cache(self):
        """
        Clear all cache in cache_dir.

        TODO: Improve cache invalidation
        """
        return self.clear_cache()

    def _get_cache_filename(self, offset: int) -> Path:
        """
        Get cache filename
        """
        return self.cache_dir / f"{offset}.parquet"

    def is_cached(self, offset: int) -> bool:
        """
        Check if the data is cached by checking if the cache file exists
        """
        return self._get_cache_filename(offset).exists()

    def _refresh(self, offset: int, data: pd.DataFrame) -> None:
        """
        Refresh cache, will write data to cache file in parquet format.
        """
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        if len(data) < self.blocksize:
            data.to_parquet(self._get_cache_filename(offset))
        elif len(data) > self.blocksize:
            for i in range(0, len(data), self.blocksize):
                data[i : i + self.blocksize].to_parquet(self._get_cache_filename(offset + i))
        else:
            data.to_parquet(self._get_cache_filename(offset))

    def load(self, offset: int, chunksize: int, data_connector: DataConnector) -> pd.DataFrame:
        """
        Load data from data_connector or cache
        """

        if chunksize % self.blocksize != 0:
            raise CacheError(
                "chunksize must be multiple of blocksize, current chunksize is {} and blocksize is {}".format(
                    chunksize, self.blocksize
                )
            )

        if chunksize != self.blocksize:
            logger.warning("chunksize must be equal to blocksize, may cause performance issue.")
        if self.is_cached(offset):
            cached_data = pd.read_parquet(self._get_cache_filename(offset))
            if len(cached_data) >= chunksize:
                return cached_data[:chunksize]
            return cached_data
        limit = max(self.blocksize, chunksize)
        data = data_connector.read(offset=offset, limit=limit)
        if data is None:
            return data

        data_list: List[pd.DataFrame] = [data]
        while len(data) < limit:
            # When generator size is less than blocksize
            # Continue to read until fit the limit
            next_data = data_connector.read(offset=offset + len(data), limit=limit - len(data))
            if next_data is None or len(next_data) == 0:
                break
            data_list.append(next_data)
        data = (
            pd.concat(
                data_list,
                ignore_index=True,
            )
            if len(data_list) > 1
            else data
        )

        self._refresh(offset, data)
        if len(data) < chunksize:
            return data
        return data[:chunksize]

    def iter(
        self, chunksize: int, data_connector: DataConnector
    ) -> Generator[pd.DataFrame, None, None]:
        """
        Load data from data_connector or cache in chunk
        """
        offset = 0
        while True:
            data = self.load(offset, chunksize, data_connector)
            if data is None or len(data) == 0:
                break
            yield data
            offset += len(data)


@hookimpl
def register(manager):
    manager.register("DiskCache", DiskCache)



from __future__ import annotations

from typing import Any

import pluggy

project_name = "sdgx.metadata.inspector"
"""
The entry-point name of this extension.

Should be used in ``pyproject.toml`` as ``[project.entry-points."{project_name}"]``
"""
hookimpl = pluggy.HookimplMarker(project_name)
"""
Hookimpl marker for this extension, extension module should use this marker

Example:

    .. code-block:: python

        @hookimpl
        def register(manager):
            ...
"""

hookspec = pluggy.HookspecMarker(project_name)


@hookspec
def register(manager):
    """
    For more information about this function, please check the :ref:`manager`

    We provided an example package for you in ``{project_root}/example/extension/dummymetadatainspector``.

    Example:

    .. code-block:: python

        class MyOwnInspector(Inspector):
            ...

        from sdgx.data_models.inspectors.extension import hookimpl

        @hookimpl
        def register(manager):
            manager.register("DummyInspector", MyOwnInspector)


    Config ``project.entry-points`` so that we can find it

    .. code-block:: toml

        [project.entry-points."sdgx.metadata.inspector"]
        {whatever-name} = "{package}.{path}.{to}.{file-with-hookimpl-function}"
    """



"""Functions to evaluate and test the performance of the RDT Transformers."""

from sdgx.models.components.sdv_rdt.performance import profiling
from sdgx.models.components.sdv_rdt.performance.performance import (
    evaluate_transformer_performance,
)

__all__ = [
    "evaluate_transformer_performance",
    "profiling",
]



import pandas as pd

from sdgx.utils import logger


class SingleTableMetric:
    """SingleTableMetric

    Metrics used to evaluate the quality of single table synthetic data.
    """

    upper_bound = None
    lower_bound = None
    metric_name = None
    metadata = None

    def __init__(self, metadata: dict) -> None:
        """Initialization

        Args:
            metadata(dict): This parameter accepts a metadata description dict, which is used to describe the column description information of the table.
        """
        self.metadata = metadata
        pass

    @classmethod
    def check_input(cls, real_data: pd.DataFrame, synthetic_data: pd.DataFrame):
        """Input check for single table input.

        Args:
            real_data(pd.DataFrame): the real (original) data table.

            synthetic_data(pd.DataFrame): the synthetic (generated) data table.
        """
        # should be pd.DataFrame
        # Input parameter must not contain None value
        if real_data is None or synthetic_data is None:
            raise TypeError("Input contains None.")

        # The data type should be same
        if type(real_data) is not type(synthetic_data):
            raise TypeError("Data type of real_data and synthetic data should be the same.")

        # if type is pd.Series, return directly
        if isinstance(real_data, pd.DataFrame):
            return real_data, synthetic_data

        # if type is not pd.Series or pd.DataFrame tranfer it to Series
        try:
            real_data = pd.DataFrame(real_data)
            synthetic_data = pd.DataFrame(synthetic_data)
            return real_data, synthetic_data
        except Exception as e:
            logger.error(f"An error occurred while converting to pd.DataFrame: {e}")

        return None, None

    def calculate(cls, real_data: pd.DataFrame, synthetic_data: pd.DataFrame):
        """Calculate the metric value between a real table and a synthetic table.

        Args:
            real_data(pd.DataFrame): the real (original) data table.

            synthetic_data(pd.DataFrame): the synthetic (generated) data table.
        """

        raise NotImplementedError()

    @classmethod
    def check_output(raw_metric_value: float):
        """Check the output value.

        Args:
            raw_metric_value (float):  the calculated raw value of the Mutual Information Similarity.
        """
        raise NotImplementedError()

    pass



from __future__ import annotations

from enum import unique
from typing import List, Union

from sdgx.models.components.sdv_rdt.transformers import (
    ClusterBasedNormalizer,
    NormalizedFrequencyEncoder,
    NormalizedLabelEncoder,
    OneHotEncoder,
)
from sdgx.models.components.utils import StrValuedBaseEnum

CategoricalEncoderInstanceType = Union[
    OneHotEncoder, NormalizedFrequencyEncoder, NormalizedLabelEncoder
]
ContinuousEncoderInstanceType = Union[ClusterBasedNormalizer]
TransformerEncoderInstanceType = Union[
    CategoricalEncoderInstanceType, ContinuousEncoderInstanceType
]


@unique
class ActivationFuncType(StrValuedBaseEnum):
    SOFTMAX = "softmax"
    TANH = "tanh"
    LINEAR = "linear"


@unique
class ColumnTransformType(StrValuedBaseEnum):
    DISCRETE = "discrete"
    CONTINUOUS = "continuous"


class SpanInfo:
    def __init__(self, dim: int, activation_fn: ActivationFuncType | str):
        self.dim: int = dim
        self.activation_fn: ActivationFuncType = ActivationFuncType(activation_fn)


class ColumnTransformInfo:
    def __init__(
        self,
        column_name: str,
        column_type: ColumnTransformType | str,
        transform: TransformerEncoderInstanceType,
        output_info: List[SpanInfo],
        output_dimensions: int,
    ):
        self.column_name: str = column_name
        self.column_type: ColumnTransformType = ColumnTransformType(column_type)
        self.transform: TransformerEncoderInstanceType = transform
        self.output_info: List[SpanInfo] = output_info
        self.output_dimensions: int = output_dimensions



"""Utils for the dataset generators."""

import numpy as np


def add_nans(array):
    """Add a random amount of NaN values to the given array.

    Args:
        array (np.array):
            1 dimensional numpy array.

    Returns:
        np.array:
            The same array with some values replaced by NaNs.
    """
    if array.dtype.kind == "i":
        array = array.astype(float)

    length = len(array)
    num_nulls = np.random.randint(1, length)
    nulls = np.random.choice(range(length), num_nulls)
    array[nulls] = np.nan
    return array



from __future__ import annotations

import os
from functools import cached_property
from typing import Callable, Generator

import pandas as pd

from sdgx.data_connectors.base import DataConnector


class GeneratorConnector(DataConnector):
    """
    A virtual data connector that wrap
    `Generator <https://docs.python.org/3/glossary.html#term-generator>`_
    into a DataConnector.

    Passing ``offset=0`` to ``read`` will reset the generator.

    Warning:
        ``offset`` and ``limit`` are ignored as ``Generator`` not supporting random access.
        But we can use :ref:`Cacher` to support it. See :ref:`Data Loader` for more details.

    Note:
        This connector is not been registered by default.
        So only be used with the library way.
    """

    @cached_property
    def identity(self) -> str:
        return f"generator-{os.getpid()}-{id(self.generator_caller)}"

    def __init__(
        self,
        generator_caller: Callable[[], Generator[pd.DataFrame, None, None]],
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.generator_caller = generator_caller
        self._generator = self.generator_caller()

    def _read(self, offset: int = 0, limit: int | None = None) -> pd.DataFrame | None:
        """
        Ingore limit and allow sequential reading.
        """
        if offset == 0:
            self._generator = self.generator_caller()

        try:
            return next(self._generator)
        except StopIteration:
            return None

    def _columns(self) -> list[str]:
        for df in self._iter():
            return list(df.columns)

    def _iter(self, offset=0, chunksize=0) -> Generator[pd.DataFrame, None, None]:
        """
        Subclass should implement this for reading data in chunk.

        See ``iter`` for more details.
        """
        return self.generator_caller()



import pytest

from sdgx.cachers.manager import CacherManager


@pytest.fixture
def manager():
    yield CacherManager()


@pytest.mark.parametrize(
    "supportd_cacher",
    ["NoCache", "DiskCache"],
)
def test_manager(supportd_cacher, manager: CacherManager):
    assert manager._normalize_name(supportd_cacher) in manager.registed_cachers


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

from typing import Any

from sdgx import data_connectors
from sdgx.data_connectors import extension
from sdgx.data_connectors.base import DataConnector
from sdgx.data_connectors.extension import project_name as PROJECT_NAME
from sdgx.manager import Manager


class DataConnectorManager(Manager):
    register_type = DataConnector
    project_name = PROJECT_NAME
    hookspecs_model = extension

    @property
    def registed_data_connectors(self):
        return self.registed_cls

    def load_all_local_model(self):
        self._load_dir(data_connectors)

    def init_data_connector(self, connector_name, **kwargs: dict[str, Any]) -> DataConnector:
        return self.init(connector_name, **kwargs)



"""Dataset Generators for Text transformers."""

from abc import ABC

import numpy as np

from sdgx.models.components.sdv_rdt.performance.datasets.base import (
    BaseDatasetGenerator,
)
from sdgx.models.components.sdv_rdt.performance.datasets.utils import add_nans


class RegexGeneratorGenerator(BaseDatasetGenerator, ABC):
    """Base class for generators that generate PII data."""

    SDTYPE = "text"


class RandomStringGenerator(RegexGeneratorGenerator):
    """Generator that creates an array of random strings."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        categories = ["Alice", "Bob", "Charlie", "Dave", "Eve"]
        return np.random.choice(a=categories, size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-05, "memory": 500.0},
            "transform": {"time": 1e-05, "memory": 500.0},
            "reverse_transform": {
                "time": 2e-05,
                "memory": 1000.0,
            },
        }


class RandomStringNaNsGenerator(RegexGeneratorGenerator):
    """Generator that creates an array of random strings with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(RandomStringGenerator.generate(num_rows).astype("O"))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-05, "memory": 400.0},
            "transform": {"time": 1e-05, "memory": 1000.0},
            "reverse_transform": {
                "time": 2e-05,
                "memory": 1000.0,
            },
        }



from __future__ import annotations

from typing import Any

from sdgx.exceptions import ManagerLoadModelError
from sdgx.manager import Manager
from sdgx.models import extension, ml, statistics
from sdgx.models.base import SynthesizerModel
from sdgx.models.extension import project_name as PROJECT_NAME


class ModelManager(Manager):
    register_type = SynthesizerModel
    project_name = PROJECT_NAME
    hookspecs_model = extension

    @property
    def registed_models(self):
        """
        redirect to registed_cls
        """

        return self.registed_cls

    def load_all_local_model(self):
        self._load_dir(ml.single_table)
        self._load_dir(ml.multi_tables)
        self._load_dir(statistics.single_table)
        self._load_dir(statistics.multi_tables)

    def init_model(self, model_name, **kwargs: dict[str, Any]) -> SynthesizerModel:
        """
        redirect to init
        """

        return self.init(model_name, **kwargs)

    def load(self, model: type[SynthesizerModel] | str, model_path, **kwargs) -> SynthesizerModel:
        if not (isinstance(model, type) or isinstance(model, str)):
            raise ManagerLoadModelError(
                "model must be type of SynthesizerModel or str for model_name"
            )
        if isinstance(model, str):
            model = self._normalize_name(model)

        if isinstance(model, str) and model not in self.registed_models:
            raise ManagerLoadModelError(f"{model} is not registered.")
        model = model if isinstance(model, type) else self.registed_models[model]

        try:
            return model.load(model_path, **kwargs)
        except Exception as e:
            raise ManagerLoadModelError(e)



from __future__ import annotations

from pathlib import Path

import pytest

from sdgx.data_connectors.csv_connector import CsvConnector
from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import CategoricalEncoderType, Metadata
from sdgx.exceptions import MetadataInvalidError


@pytest.fixture
def dataloader(demo_single_table_path, cacher_kwargs):
    d = DataLoader(CsvConnector(path=demo_single_table_path), cacher_kwargs=cacher_kwargs)
    yield d
    d.finalize(clear_cache=True)


@pytest.fixture
def metadata(dataloader):
    yield Metadata.from_dataloader(dataloader)


def test_metadata(metadata: Metadata):
    assert metadata.discrete_columns == metadata.get("discrete_columns")
    assert metadata.id_columns == metadata.get("id_columns")
    assert metadata.datetime_columns == metadata.get("datetime_columns")
    assert metadata.bool_columns == metadata.get("bool_columns")
    assert metadata.int_columns == metadata.get("int_columns")
    assert metadata.float_columns == metadata.get("float_columns")

    metadata.set("a", "something")
    assert metadata.get("a") == {"something"}

    assert metadata._dump_json()


def test_change_metadata(metadata: Metadata):
    metadata = metadata.model_copy()
    col = "age"
    assert col in metadata.int_columns
    assert col not in metadata.datetime_columns
    metadata.change_column_type(col, "int", "datetime")
    assert col in metadata.datetime_columns
    assert col not in metadata.int_columns
    metadata.change_column_type(col, "datetime", "int")
    assert col in metadata.int_columns
    assert col not in metadata.datetime_columns


def test_remove_metadata(metadata: Metadata):
    metadata = metadata.model_copy()
    col = "age"
    assert col in metadata.int_columns
    metadata.remove_column([col])
    assert col not in metadata.int_columns


def test_metadata_save_load(metadata: Metadata, tmp_path: Path):
    test_path = tmp_path / "metadata_path_test.json"
    metadata.save(test_path)
    # load from path
    new_meatadata = Metadata.load(test_path)
    assert metadata == new_meatadata


def test_metadata_primary_key(metadata: Metadata):
    # inspect fnlwgt to ID type (for test)
    metadata.add("id_columns", "fnlwgt")
    # set fnlwgt as primary key
    metadata.update_primary_key(["fnlwgt"])
    assert metadata.primary_keys == {"fnlwgt"}


def test_metadata_primary_query_filed_tags():
    metadata = Metadata()
    metadata.set("id_columns", {"user_id"})
    metadata.set("int_columns", {"user_id", "age"})

    # query the "user_id" field tags
    results = metadata.query("user_id")
    results_list = list(results)
    print(results_list)
    assert "id_columns" in results_list
    assert "int_columns" in results_list


def test_metadata_check(metadata: Metadata):
    # For the example table, it does not contain a primary key
    # but it can be used in  single-table data synthetic tasks.
    # clear primary key
    metadata.update_primary_key([])
    # do meatadata check
    metadata.check()


def test_demo_multi_table_data_metadata_parent(demo_multi_data_parent_matadata):
    # self check
    demo_multi_data_parent_matadata.check()
    # check each col's data type
    assert demo_multi_data_parent_matadata.get_column_data_type("Store") == "id"
    assert demo_multi_data_parent_matadata.get_column_data_type("StoreType") == "discrete"
    assert demo_multi_data_parent_matadata.get_column_data_type("Assortment") == "discrete"
    assert demo_multi_data_parent_matadata.get_column_data_type("CompetitionDistance") == "int"
    assert (
        demo_multi_data_parent_matadata.get_column_data_type("CompetitionOpenSinceMonth") == "int"
    )
    assert demo_multi_data_parent_matadata.get_column_data_type("Promo2") == "int"
    assert demo_multi_data_parent_matadata.get_column_data_type("Promo2SinceWeek") == "int"
    assert demo_multi_data_parent_matadata.get_column_data_type("Promo2SinceYear") == "int"
    assert demo_multi_data_parent_matadata.get_column_data_type("PromoInterval") == "discrete"
    # check pii
    for each_col in demo_multi_data_parent_matadata.column_list:
        assert demo_multi_data_parent_matadata.get_column_pii(each_col) is False
    assert len(demo_multi_data_parent_matadata.pii_columns) is 0
    # check dump
    assert "column_data_type" in demo_multi_data_parent_matadata.dump().keys()


def test_demo_multi_table_data_metadata_child(demo_multi_data_child_matadata):
    # self check
    demo_multi_data_child_matadata.check()
    # check each col's data type
    assert demo_multi_data_child_matadata.get_column_data_type("Store") == "int"
    assert demo_multi_data_child_matadata.get_column_data_type("Date") == "datetime"
    assert demo_multi_data_child_matadata.get_column_data_type("Customers") == "int"
    assert demo_multi_data_child_matadata.get_column_data_type("StateHoliday") == "const"
    assert demo_multi_data_child_matadata.get_column_data_type("Sales") == "int"
    assert demo_multi_data_child_matadata.get_column_data_type("Promo") == "int"
    assert demo_multi_data_child_matadata.get_column_data_type("DayOfWeek") == "int"
    assert demo_multi_data_child_matadata.get_column_data_type("Open") == "int"
    assert demo_multi_data_child_matadata.get_column_data_type("SchoolHoliday") == "int"
    # check pii
    for each_col in demo_multi_data_child_matadata.column_list:
        assert demo_multi_data_child_matadata.get_column_pii(each_col) is False
    assert len(demo_multi_data_child_matadata.pii_columns) is 0
    # check dump
    assert "column_data_type" in demo_multi_data_child_matadata.dump().keys()


def test_meta_encoder(metadata: Metadata):
    metadata = metadata.model_copy()
    metadata.categorical_threshold = {1: "aaa"}
    with pytest.raises(MetadataInvalidError):
        metadata.check()
    metadata.categorical_threshold[1] = CategoricalEncoderType.ONEHOT
    metadata.check()
    metadata.categorical_threshold["1"] = CategoricalEncoderType.ONEHOT
    with pytest.raises(MetadataInvalidError):
        metadata.check()
    del metadata.categorical_threshold["1"]
    assert "education" in metadata.discrete_columns
    metadata.categorical_encoder = {
        "education": CategoricalEncoderType.ONEHOT,
    }
    metadata.check()
    metadata.categorical_encoder["1"] = CategoricalEncoderType.ONEHOT
    with pytest.raises(MetadataInvalidError):
        metadata.check()
    del metadata.categorical_encoder["1"]
    metadata.categorical_encoder["1"] = "a"
    with pytest.raises(MetadataInvalidError):
        metadata.check()


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import pytest

from sdgx.data_processors.manager import DataProcessorManager


@pytest.fixture
def manager():
    yield DataProcessorManager()


@pytest.mark.parametrize(
    "supported_data_processor",
    [],
)
def test_manager(supported_data_processor, manager: DataProcessorManager):
    assert manager._normalize_name(supported_data_processor) in manager.registed_data_processors


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

import pytest

from sdgx.data_models.metadata import Metadata
from sdgx.data_models.relationship import KeyTuple, Relationship
from sdgx.exceptions import RelationshipInitError

parent_metadata = Metadata(
    primary_keys=["parent_id"], column_list=["parent_id"], id_columns={"parent_id"}
)
error_parent_metadata = Metadata(
    primary_keys=["parent_id"], column_list=["parent_id"], id_columns={"foo"}
)
child_metadata = Metadata(
    primary_keys=["child_id"],
    column_list=["parent_id", "child_id"],
    id_columns={"parent_id", "child_id"},
)
error_child_metadata = Metadata(
    primary_keys=["child_id"], column_list=["parent_id", "child_id"], id_columns={"child_id"}
)


@pytest.mark.parametrize(
    "parent_table, parent_metadata, child_table, child_metadata, foreign_keys, exception",
    [
        (
            "parent",
            parent_metadata,
            "child",
            child_metadata,
            [KeyTuple("parent_id", "parent_id")],
            None,
        ),
        (
            "parent",
            error_parent_metadata,
            "child",
            child_metadata,
            [KeyTuple("parent_id", "parent_id")],
            RelationshipInitError,
        ),
        (
            "parent",
            parent_metadata,
            "child",
            error_child_metadata,
            [KeyTuple("parent_id", "parent_id")],
            RelationshipInitError,
        ),
        (
            "parent",
            parent_metadata,
            "parent",
            parent_metadata,
            [KeyTuple("parent_id", "parent_id")],
            RelationshipInitError,
        ),
        ("parent", parent_metadata, "parent", parent_metadata, [], RelationshipInitError),
        (
            "",
            parent_metadata,
            "child",
            child_metadata,
            [KeyTuple("parent_id", "parent_id")],
            RelationshipInitError,
        ),
        (
            "parent",
            parent_metadata,
            "",
            child_metadata,
            [KeyTuple("parent_id", "parent_id")],
            RelationshipInitError,
        ),
        (
            "",
            parent_metadata,
            "",
            child_metadata,
            [KeyTuple("parent_id", "parent_id")],
            RelationshipInitError,
        ),
        ("", parent_metadata, "", child_metadata, [], RelationshipInitError),
    ],
)
def test_build(parent_table, parent_metadata, child_table, child_metadata, foreign_keys, exception):
    if exception:
        with pytest.raises(exception):
            Relationship.build(
                parent_table=parent_table,
                parent_metadata=parent_metadata,
                child_table=child_table,
                child_metadata=child_metadata,
                foreign_keys=foreign_keys,
            )
    else:
        relationship = Relationship.build(
            parent_table=parent_table,
            parent_metadata=parent_metadata,
            child_table=child_table,
            child_metadata=child_metadata,
            foreign_keys=foreign_keys,
        )

        assert relationship.parent_table == parent_table
        assert relationship.child_table == child_table
        assert relationship.foreign_keys == foreign_keys


def test_save_and_load(tmpdir):
    save_file = tmpdir / "relationship.json"
    relationship = Relationship.build(
        parent_table="parent",
        parent_metadata=Metadata(
            primary_keys=["parent_id"], column_list=["parent_id"], id_columns={"parent_id"}
        ),
        child_table="child",
        child_metadata=Metadata(
            primary_keys=["child_id"],
            column_list=["parent_id", "child_id"],
            id_columns={"parent_id", "child_id"},
        ),
        foreign_keys=[KeyTuple("parent_id", "parent_id")],
    )

    relationship.save(save_file)
    assert save_file.exists()

    assert relationship == Relationship.load(save_file)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

import pandas as pd
import pytest

from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.exceptions import InitializationError
from sdgx.models.LLM.single_table.gpt import SingleTableGPTModel


@pytest.fixture
def raw_data(demo_single_table_path):
    yield pd.read_csv(demo_single_table_path).head(100)


@pytest.fixture
def single_table_gpt_model():
    yield SingleTableGPTModel()


# When reading the code, please collapse this list
# COLLAPSE ME
gpt_response_list = [
    """
Here are 20 similar data entries generated based on the provided information:

sample 21: relationship is Husband, fnlwgt is 145441, educational-num is 9, education is HS-grad, occupation is Exec-managerial, gender is Male, race is White, workclass is Private, capital-gain is 0, native-country is United-States, marital-status is Married-civ-spouse, income is >50K, age is 40, capital-loss is 1485, hours-per-week is 40
sample 22: income is <=50K, gender is Male, education is Assoc-acdm, native-country is ?, educational-num is 12, hours-per-week is 7, occupation is Prof-specialty, capital-gain is 0, capital-loss is 0, fnlwgt is 154164, race is White, workclass is Private, age is 66, relationship is Not-in-family, marital-status is Never-married
sample 23: relationship is Own-child, marital-status is Never-married, occupation is Transport-moving, income is <=50K, race is White, workclass is Private, hours-per-week is 30, capital-gain is 0, native-country is United-States, capital-loss is 0, fnlwgt is 283499, gender is Male, education is Some-college, age is 20, educational-num is 10
sample 24: marital-status is Married-civ-spouse, gender is Male, capital-gain is 0, capital-loss is 0, educational-num is 9, hours-per-week is 40, fnlwgt is 170772, income is <=50K, occupation is Other-service, relationship is Husband, race is White, education is HS-grad, age is 30, native-country is United-States, workclass is Local-gov
sample 25: native-country is United-States, income is <=50K, capital-gain is 0, hours-per-week is 40, fnlwgt is 367306, educational-num is 10, capital-loss is 0, race is White, relationship is Own-child, gender is Female, occupation is Tech-support, workclass is Private, age is 25, marital-status is Never-married, education is Some-college
sample 26: occupation is Exec-managerial, fnlwgt is 81973, native-country is United-States, capital-loss is 0, educational-num is 14, relationship is Husband, marital-status is Married-civ-spouse, workclass is Private, capital-gain is 0, gender is Male, race is Asian-Pac-Islander, hours-per-week is 40, age is 59, income is >50K, education is Masters
sample 27: gender is Male, fnlwgt is 287268, workclass is Private, native-country is United-States, capital-loss is 0, age is 28, educational-num is 10, hours-per-week is 35, income is <=50K, relationship is Not-in-family, occupation is Other-service, capital-gain is 0, race is White, education is Some-college, marital-status is Never-married
sample 28: race is White, income is >50K, native-country is United-States, workclass is Private, capital-loss is 0, occupation is Craft-repair, hours-per-week is 60, educational-num is 13, gender is Male, fnlwgt is 176729, relationship is Husband, marital-status is Married-civ-spouse, capital-gain is 0, age is 25, education is Bachelors
sample 29: capital-loss is 0, native-country is Cambodia, educational-num is 9, marital-status is Married-civ-spouse, workclass is Self-emp-not-inc, education is HS-grad, age is 42, occupation is Farming-fishing, race is Asian-Pac-Islander, income is >50K, gender is Male, hours-per-week is 40, fnlwgt is 303044, capital-gain is 0, relationship is Husband
sample 30: age is 17, capital-loss is 0, relationship is Own-child, educational-num is 6, workclass is Private, income is <=50K, native-country is United-States, hours-per-week is 15, race is White, occupation is Other-service, fnlwgt is 202521, education is 10th, gender is Male, capital-gain is 0, marital-status is Never-married
sample 31: fnlwgt is 173858, occupation is Exec-managerial, education is Bachelors, native-country is China, hours-per-week is 35, marital-status is Married-civ-spouse, relationship is Husband, age is 38, capital-gain is 7688, workclass is Private, race is Asian-Pac-Islander, income is >50K, capital-loss is 0, gender is Male, educational-num is 13
sample 32: age is 74, workclass is Private, marital-status is Widowed, income is <=50K, hours-per-week is 40, occupation is Priv-house-serv, education is Assoc-voc, gender is Female, race is White, relationship is Not-in-family, fnlwgt is 68326, capital-loss is 0, native-country is United-States, capital-gain is 0, educational-num is 11
sample 33: gender is Male, race is White, age is 28, capital-gain is 0, marital-status is Married-civ-spouse, hours-per-week is 40, relationship is Husband, education is HS-grad, educational-num is 9, income is <=50K, workclass is Private, capital-loss is 0, fnlwgt is 66095, occupation is Sales, native-country is United-States
sample 34: education is Bachelors, occupation is Adm-clerical, race is White, marital-status is Married-civ-spouse, capital-gain is 7688, gender is Male, income is >50K, relationship is Husband, workclass is Private, fnlwgt is 206814, age is 58, hours-per-week is 50, capital-loss is 0, native-country is United-States, educational-num is 13
sample 35: age is 44, occupation is Craft-repair, capital-loss is 0, workclass is Federal-gov, income is >50K, fnlwgt is 243636, capital-gain is 0, education is Assoc-voc, relationship is Husband, educational-num is 11, race is White, marital-status is Married-civ-spouse, hours-per-week is 40, native-country is United-States, gender is Male
sample 36: education is Masters, fnlwgt is 37070, marital-status is Married-civ-spouse, age is 33, relationship is Husband, capital-gain is 0, educational-num is 14, workclass is State-gov, gender is Male, income is <=50K, occupation is Prof-specialty, capital-loss is 0, race is White, native-country is Canada, hours-per-week is 60
sample 37: workclass is Private, educational-num is 7, education is 11th, income is <=50K, relationship is Not-in-family, gender is Male, hours-per-week is 40, native-country is Puerto-Rico, age is 23, fnlwgt is 224217, occupation is Transport-moving, capital-gain is 0, marital-status is Never-married, capital-loss is 0, race is White
sample 38: hours-per-week is 40, capital-gain is 0, gender is Male, marital-status is Never-married, age is 25, capital-loss is 0, native-country is ?, fnlwgt is 310864, income is <=50K, educational-num is 13, race is Black, workclass is Private, education is Bachelors, relationship is Not-in-family, occupation is Tech-support
sample 39: capital-gain is 0, hours-per-week is 55, capital-loss is 0, age is 30, income is <=50K, education is Bachelors, relationship is Not-in-family, marital-status is Never-married, occupation is Exec-managerial, native-country is United-States, gender is Female, race is White, educational-num is 13, workclass is Private, fnlwgt is 128016
sample 40: hours-per-week is 40, fnlwgt is 174515, capital-loss is 0, marital-status is Widowed, native-country is United-States, capital-gain is 0, age is 40, education is HS-grad, occupation is Machine-op-inspct, educational-num is 9, relationship is Unmarried, race is White, gender is Female, workclass is Private, income is <=50K
""",
    """Based on the provided information, here are 15 synthetic data samples generated:

Sample 0: income is <=50K, race is White, marital-status is Married-civ-spouse, gender is Male, age is 36, workclass is Self-emp-inc, education is HS-grad, relationship is Husband, native-country is United-States, educational-num is 9.0, occupation is Farming-fishing, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 80.0, fnlwgt is 48063

Sample 1: income is <=50K, race is White, marital-status is Never-married, gender is Male, age is 34, workclass is Private, education is Masters, relationship is Not-in-family, native-country is United-States, educational-num is 14.0, occupation is Prof-specialty, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 40.0, fnlwgt is 189759

Sample 2: income is <=50K, race is Amer-Indian-Eskimo, marital-status is Never-married, gender is Female, age is 19, workclass is Private, education is HS-grad, relationship is Own-child, native-country is United-States, educational-num is 9.0, occupation is Other-service, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 35.0, fnlwgt is 106183

Sample 3: income is <=50K, race is White, marital-status is Married-civ-spouse, gender is Male, age is 42, workclass is Private, education is HS-grad, relationship is Husband, native-country is United-States, educational-num is 9.0, occupation is Craft-repair, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 45.0, fnlwgt is 190910

Sample 4: income is >50K, race is White, marital-status is Married-civ-spouse, gender is Male, age is 59, workclass is Private, education is 7th-8th, relationship is Husband, native-country is United-States, educational-num is 4.0, occupation is Craft-repair, capital-loss is 0.0, capital-gain is 5178.0, hours-per-week is 50.0, fnlwgt is 107318

Sample 5: income is >50K, race is White, marital-status is Divorced, gender is Male, age is 51, workclass is Private, education is Some-college, relationship is Not-in-family, native-country is United-States, educational-num is 10.0, occupation is Craft-repair, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 42.0, fnlwgt is 43354

Sample 6: income is <=50K, race is White, marital-status is Separated, gender is Male, age is 46, workclass is Private, education is HS-grad, relationship is Not-in-family, native-country is United-States, educational-num is 9.0, occupation is Transport-moving, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 40.0, fnlwgt is 170338

Sample 7: income is >50K, race is White, marital-status is Married-civ-spouse, gender is Male, age is 45, workclass is Self-emp-not-inc, education is Some-college, relationship is Husband, native-country is United-States, educational-num is 10.0, occupation is Sales, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 60.0, fnlwgt is 355978

Sample 8: income is <=50K, race is White, marital-status is Never-married, gender is Male, age is 33, workclass is Private, education is Bachelors, relationship is Not-in-family, native-country is United-States, educational-num is 13.0, occupation is Sales, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 45.0, fnlwgt is 90409

Sample 9: income is <=50K, race is White, marital-status is Married-civ-spouse, gender is Male, age is 29, workclass is Private, education is 11th, relationship is Husband, native-country is United-States, educational-num is 7.0, occupation is Other-service, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 40.0, fnlwgt is 103634

Sample 10: income is >50K, race is White, marital-status is Never-married, gender is Male, age is 22, workclass is Private, education is HS-grad, relationship is Not-in-family, native-country is United-States, educational-num is 9.0, occupation is Other-service, capital-loss is 0.0, capital-gain is 14084.0, hours-per-week is 60.0, fnlwgt is 54164

Sample 11: income is <=50K, race is White, marital-status is Never-married, gender is Male, age is 17, workclass is ?, education is 10th, relationship is Own-child, native-country is United-States, educational-num is 6.0, occupation is ?, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 40.0, fnlwgt is 165361

Sample 12: income is <=50K, race is White, marital-status is Married-civ-spouse, gender is Male, age is 23, workclass is Private, education is 10th, relationship is Husband, native-country is United-States, educational-num is 6.0, occupation is Farming-fishing, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 40.0, fnlwgt is 306309

Sample 13: income is <=50K, race is White, marital-status is Married-civ-spouse, gender is Male, age is 38, workclass is Private, education is HS-grad, relationship is Husband, native-country is United-States, educational-num is 9.0, occupation is Farming-fishing, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 50.0, fnlwgt is 89814

Sample 14: income is >50K, race is White, marital-status is Married-civ-spouse, gender is Male, age is 36, workclass is Local-gov, education is Bachelors, relationship is Husband, native-country is United-States, educational-num is 13.0, occupation is Prof-specialty, capital-loss is 0.0, capital-gain is 0.0, hours-per-week is 40.0, fnlwgt is 403681

""",
    """
marital-status is Married-civ-spouse, relationship is Husband, income is <=50K, age is 45, fnlwgt is 200000, occupation is Exec-managerial, native-country is United-States, hours-per-week is 40, workclass is Private, gender is Male, educational-num is 13, capital-gain is 0, capital-loss is 0, race is White, education is Bachelors
marital-status is Never-married, relationship is Not-in-family, income is <=50K, age is 25, fnlwgt is 150000, occupation is Sales, native-country is United-States, hours-per-week is 35, workclass is Private, gender is Female, educational-num is 10, capital-gain is 0, capital-loss is 0, race is White, education is Some-college
marital-status is Divorced, relationship is Unmarried, income is <=50K, age is 35, fnlwgt is 180000, occupation is Adm-clerical, native-country is United-States, hours-per-week is 30, workclass is Private, gender is Female, educational-num is 12, capital-gain is 0, capital-loss is 0, race is White, education is HS-grad
marital-status is Married-civ-spouse, relationship is Wife, income is >50K, age is 50, fnlwgt is 250000, occupation is Prof-specialty, native-country is United-States, hours-per-week is 45, workclass is Self-emp-not-inc, gender is Female, educational-num is 14, capital-gain is 5000, capital-loss is 0, race is White, education is Masters
marital-status is Married-civ-spouse, relationship is Husband, income is >50K, age is 40, fnlwgt is 220000, occupation is Exec-managerial, native-country is United-States, hours-per-week is 50, workclass is Private, gender is Male, educational-num is 13, capital-gain is 10000, capital-loss is 0, race is White, education is Bachelors
marital-status is Never-married, relationship is Own-child, income is <=50K, age is 18, fnlwgt is 100000, occupation is Other-service, native-country is United-States, hours-per-week is 20, workclass is Private, gender is Male, educational-num is 9, capital-gain is 0, capital-loss is 0, race is Black, education is Some-college
marital-status is Married-civ-spouse, relationship is Wife, income is >50K, age is 35, fnlwgt is 180000, occupation is Prof-specialty, native-country is United-States, hours-per-week is 40, workclass is Private, gender is Female, educational-num is 14, capital-gain is 8000, capital-loss is 0, race is White, education is Masters
marital-status is Divorced, relationship is Unmarried, income is <=50K, age is 30, fnlwgt is 160000, occupation is Adm-clerical, native-country is United-States, hours-per-week is 35, workclass is Private, gender is Female, educational-num is 12, capital-gain is 0, capital-loss is 0, race is White, education is HS-grad
marital-status is Married-civ-spouse, relationship is Husband, income is >50K, age is 55, fnlwgt is 280000, occupation is Exec-managerial, native-country is United-States, hours-per-week is 60, workclass is Private, gender is Male, educational-num is 13, capital-gain is 15000, capital-loss is 0, race is White, education is Bachelors
marital-status is Never-married, relationship is Not-in-family, income is <=50K, age is 28, fnlwgt is 140000, occupation is Sales, native-country is United-States, hours-per-week is 40, workclass is Private, gender is Female, educational-num is 10, capital-gain is 0, capital-loss is 0, race is White, education is Some-college
marital-status is Divorced, relationship is Unmarried, income is <=50K, age is 40, fnlwgt is 200000, occupation is Adm-clerical, native-country is United-States, hours-per-week is 30, workclass is Private, gender is Female, educational-num is 12, capital-gain is 0, capital-loss is 0, race is White, education is HS-grad
marital-status is Married-civ-spouse, relationship is Wife, income is >50K, age is 45, fnlwgt is 220000, occupation is Prof-specialty, native-country is United-States, hours-per-week is 50, workclass is Self-emp-not-inc, gender is Female, educational-num is 14, capital-gain is 10000, capital-loss is 0, race is White, education is Masters
marital-status is Married-civ-spouse, relationship is Husband, income is >50K, age is 38, fnlwgt is 180000, occupation is Exec-managerial, native-country is United-States, hours-per-week is 45, workclass is Private, gender is Male, educational-num is 13, capital-gain is 8000, capital-loss is 0, race is White, education is Bachelors
marital-status is Never-married, relationship is Own-child, income is <=50K, age is 20, fnlwgt is 120000, occupation is Other-service, native-country is United-States, hours-per-week is 20, workclass is Private, gender is Male, educational-num is 9, capital-gain is 0, capital-loss is 0, race is Black, education is Some-college
marital-status is Married-civ-spouse, relationship is Wife, income is >50K, age is 30, fnlwgt is 160000, occupation is Prof-specialty, native-country is United-States, hours-per-week is 40, workclass is Private, gender is Female, educational-num is 14, capital-gain is 5000, capital-loss is 0, race is White, education is Masters
marital-status is Divorced, relationship is Unmarried, income is <=50K, age is 25, fnlwgt is 140000, occupation is Adm-clerical, native-country is United-States, hours-per-week is 35, workclass is Private, gender is Female, educational-num is 12, capital-gain is 0, capital-loss is 0, race is White, education is HS-grad
marital-status is Married-civ-spouse, relationship is Husband, income is >50K, age is 50, fnlwgt is 250000, occupation is Exec-managerial, native-country is United-States, hours-per-week is 60, workclass is Private, gender is Male, educational-num is 13, capital-gain is 10000, capital-loss is 0, race is White, education is Bachelors
marital-status is Never-married, relationship is Not-in-family, income is <=50K, age is 30, fnlwgt is 160000, occupation is Sales, native-country is United-States, hours-per-week is 40, workclass is Private, gender is Female, educational-num is 10, capital-gain is 0, capital-loss is 0, race is White, education is Some-college
marital-status is Divorced, relationship is Unmarried, income is <=50K, age is 35, fnlwgt is 180000, occupation is Adm-clerical, native-country is United-States, hours-per-week is 30, workclass is Private, gender is Female, educational-num is 12, capital-gain is 0, capital-loss is 0, race is White, education is HS-grad
marital-status is Married-civ-spouse, relationship is Wife, income is >50K, age is 55, fnlwgt is 280000, occupation is Prof-specialty, native-country is United-States, hours-per-week is 50, workclass is Private, gender is Female, educational-num is 14, capital-gain is 15000, capital-loss is 0, race is White, education is Masters

""",
    """marital-status is Married-civ-spouse, capital-gain is 0, occupation is Exec-managerial, education is Bachelors, fnlwgt is 189778, age is 35, relationship is Husband, hours-per-week is 40, income is <=50K, native-country is United-States, gender is Male, capital-loss is 0, race is White, educational-num is 13, workclass is Private, has_car is True
marital-status is Never-married, capital-gain is 0, occupation is Adm-clerical, education is HS-grad, fnlwgt is 183934, age is 28, relationship is Not-in-family, hours-per-week is 35, income is <=50K, native-country is United-States, gender is Female, capital-loss is 0, race is White, educational-num is 9, workclass is Private, has_car is False
marital-status is Divorced, capital-gain is 0, occupation is Handlers-cleaners, education is 11th, fnlwgt is 234721, age is 42, relationship is Unmarried, hours-per-week is 45, income is <=50K, native-country is United-States, gender is Male, capital-loss is 0, race is Black, educational-num is 7, workclass is Private, has_car is True
marital-status is Married-civ-spouse, capital-gain is 0, occupation is Prof-specialty, education is Masters, fnlwgt is 216129, age is 41, relationship is Husband, hours-per-week is 50, income is >50K, native-country is United-States, gender is Male, capital-loss is 0, race is White, educational-num is 14, workclass is Self-emp-inc, has_car is True
marital-status is Married-civ-spouse, capital-gain is 0, occupation is Craft-repair, education is Some-college, fnlwgt is 112497, age is 52, relationship is Husband, hours-per-week is 60, income is >50K, native-country is United-States, gender is Male, capital-loss is 0, race is White, educational-num is 10, workclass is Private, has_car is True""",
    """
marital-status = Married-civ-spouse, capital-gain = 0, occupation = Exec-managerial, education = Bachelors, fnlwgt = 189778, age = 35, relationship = Husband, hours-per-week = 40, income = <=50K, native-country = United-States, gender = Male, capital-loss = 0, race = White, educational-num = 13, workclass = Private, has_car = True
marital-status = Never-married, capital-gain = 0, occupation = Adm-clerical, education = HS-grad, fnlwgt = 183934, age = 28, relationship = Not-in-family, hours-per-week = 35, income = <=50K, native-country = United-States, gender = Female, capital-loss = 0, race = White, educational-num = 9, workclass = Private, has_car = False
marital-status = Divorced, capital-gain = 0, occupation = Handlers-cleaners, education = 11th, fnlwgt = 234721, age = 42, relationship = Unmarried, hours-per-week = 45, income = <=50K, native-country = United-States, gender = Male, capital-loss = 0, race = Black, educational-num = 7, workclass = Private, has_car = True
marital-status = Married-civ-spouse, capital-gain = 0, occupation = Prof-specialty, education = Masters, fnlwgt = 216129, age = 41, relationship = Husband, hours-per-week = 50, income = >50K, native-country = United-States, gender = Male, capital-loss = 0, race = White, educational-num = 14, workclass = Self-emp-inc, has_car = True
marital-status = Married-civ-spouse, capital-gain = 0, occupation = Craft-repair, education = Some-college, fnlwgt = 112497, age = 52, relationship = Husband, hours-per-week = 60, income = >50K, native-country = United-States, gender = Male, capital-loss = 0, race = White, educational-num = 10, workclass = Private, has_car = True
""",
]

gpt_response_sample_count = [20, 15, 20, 5, 5]


def test_singletable_gpt_model_openapi_setting(single_table_gpt_model: SingleTableGPTModel):
    open_ai_key = "sk-XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    open_ai_base = "https://api.mock.openai.base.com"
    open_ai_model = "gpt-4o-mini"
    single_table_gpt_model.set_openAI_settings(open_ai_base, open_ai_key)
    single_table_gpt_model.gpt_model = open_ai_model
    client = single_table_gpt_model.openai_client()
    assert client.base_url == open_ai_base
    assert client.api_key == open_ai_key
    assert single_table_gpt_model.gpt_model == open_ai_model


def test_singletable_gpt_model(
    single_table_gpt_model: SingleTableGPTModel,
    raw_data: pd.DataFrame,
    demo_single_table_data_loader: DataLoader,
):
    single_table_gpt_model.fit(raw_data)
    assert single_table_gpt_model.columns == [
        "age",
        "workclass",
        "fnlwgt",
        "education",
        "educational-num",
        "marital-status",
        "occupation",
        "relationship",
        "race",
        "gender",
        "capital-gain",
        "capital-loss",
        "hours-per-week",
        "native-country",
        "income",
    ]
    assert single_table_gpt_model.openai_API_url == "https://api.openai.com/v1/"
    # the key is not set
    assert not single_table_gpt_model.openai_API_key
    assert single_table_gpt_model.max_tokens == 4000
    assert single_table_gpt_model.temperature == 0.1
    assert single_table_gpt_model.timeout == 90
    assert "gpt-3.5" in single_table_gpt_model.gpt_model.lower()
    assert single_table_gpt_model.use_raw_data is True
    assert single_table_gpt_model.use_dataloader is False
    assert single_table_gpt_model.use_metadata is False
    assert single_table_gpt_model.query_batch == 30
    assert not single_table_gpt_model.off_table_features
    assert len(single_table_gpt_model.columns) > 0
    # train with dataloader
    single_table_gpt_model.fit(demo_single_table_data_loader)


@pytest.mark.parametrize("response_index", range(len(gpt_response_list)))
def test_feature_extraction_data(
    response_index: int, single_table_gpt_model: SingleTableGPTModel, raw_data: pd.DataFrame
):
    single_table_gpt_model.fit(raw_data)
    response_content = gpt_response_list[response_index]
    res = single_table_gpt_model.extract_samples_from_response(response_content)
    assert type(res) is list
    # assert shape of extracted features
    assert len(res) == gpt_response_sample_count[response_index]
    assert len(res[0]) == len(single_table_gpt_model.columns)
    res_df = pd.DataFrame(
        res, columns=single_table_gpt_model.columns + single_table_gpt_model.off_table_features
    )
    assert res_df.shape == (
        gpt_response_sample_count[response_index],
        len(single_table_gpt_model.columns),
    )
    sample_list = single_table_gpt_model._sample_lines
    message = single_table_gpt_model._form_message_with_data(sample_list, 20)
    assert type(message) is str
    for each_col in raw_data.columns:
        assert each_col in message
    assert type(sample_list) is list
    assert len(sample_list) == len(raw_data)
    fake_openAI_KEY = "sk-qXCXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
    occur_error = False
    try:
        single_table_gpt_model.check()
        occur_error = False
    except Exception as e:
        occur_error = True
        assert type(e) is InitializationError
    assert occur_error is True
    # set and check again
    single_table_gpt_model.set_openAI_settings("https://api.openai.com/v1/", fake_openAI_KEY)
    single_table_gpt_model.check()


@pytest.mark.parametrize("response_index", range(len(gpt_response_list)))
def test_feature_extraction_metadata(
    response_index: int,
    single_table_gpt_model: SingleTableGPTModel,
    demo_single_table_metadata: Metadata,
):
    single_table_gpt_model.fit(demo_single_table_metadata)
    single_table_gpt_model.off_table_features = ["has_car"]
    response_content = gpt_response_list[response_index]
    res = single_table_gpt_model.extract_samples_from_response(response_content)
    assert len(res) == gpt_response_sample_count[response_index]
    assert len(res[0]) == len(single_table_gpt_model.columns) + len(
        single_table_gpt_model.off_table_features
    )
    res_df = pd.DataFrame(
        res, columns=single_table_gpt_model.columns + single_table_gpt_model.off_table_features
    )
    assert res_df.shape == (
        gpt_response_sample_count[response_index],
        len(single_table_gpt_model.columns) + len(single_table_gpt_model.off_table_features),
    )
    message = single_table_gpt_model._form_message_with_metadata(20)
    for each_col in demo_single_table_metadata.column_list:
        assert each_col in message
    assert type(message) is str
    # train with metadata with another way
    single_table_gpt_model.fit(metadata=demo_single_table_metadata)



import json

import pytest

from sdgx.cli.message import ExceptionMessage, NormalMessage
from sdgx.exceptions import SdgxError


@pytest.mark.parametrize("return_val", [0, "123", [1, 2, 3], {"a": 1, "b": 2}])
def test_normal_message(return_val):
    NormalMessage.from_return_val(return_val)._dump_json == json.dumps(
        {
            "code": 0,
            "msg": "Success",
            "payload": return_val if isinstance(return_val, dict) else {"return_val": return_val},
        }
    )


def unknown_exception():
    raise Exception


def sdgx_exception():
    raise SdgxError


@pytest.mark.parametrize("exception_caller", [unknown_exception, sdgx_exception])
def test_exception_message(exception_caller):
    try:
        exception_caller()
    except Exception as e:
        msg = ExceptionMessage.from_exception(e)
        assert msg._dump_json()
        assert msg.code != 0
        assert msg.payload
        assert "details" in msg.payload


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

from pathlib import Path

import pandas as pd

from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.exceptions import SynthesizerInitError


class SynthesizerModel:
    use_dataloader: bool = False
    use_raw_data: bool = False

    def __init__(self, *args, **kwargs) -> None:
        # specify data access type
        if "use_dataloader" in kwargs.keys():
            self.use_dataloader = kwargs["use_dataloader"]
        if "use_raw_data" in kwargs.keys():
            self.use_raw_data = kwargs["use_raw_data"]

    def _check_access_type(self):
        if self.use_dataloader == self.use_raw_data == False:
            raise SynthesizerInitError(
                "Data access type not specified, please use `use_raw_data: bool` or `use_dataloader: bool` to specify data access type."
            )
        elif self.use_dataloader == self.use_raw_data == True:
            raise SynthesizerInitError("Duplicate data access type found.")

    def fit(self, metadata: Metadata, dataloader: DataLoader, *args, **kwargs):
        """
        Fit the model using the given metadata and dataloader.

        Args:
            metadata (Metadata): The metadata to use.
            dataloader (DataLoader): The dataloader to use.
        """
        raise NotImplementedError

    def sample(self, count: int, *args, **kwargs) -> pd.DataFrame:
        """
        Sample data from the model.

        Args:
            count (int): The number of samples to generate.

        Returns:
            pd.DataFrame: The generated data.
        """

        raise NotImplementedError

    def save(self, save_dir: str | Path):
        """
        Dump model to file.

        Args:
            save_dir (str | Path): The directory to save the model.
        """

        raise NotImplementedError

    @classmethod
    def load(cls, save_dir: str | Path, **kwargs) -> "SynthesizerModel":
        """
        Load model from file.

        Args:
            save_dir (str | Path): The directory to load the model from.
        """
        raise NotImplementedError



import copy

import numpy as np
import pandas as pd
import pytest

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.transformers.const import ConstValueTransformer


@pytest.fixture
def test_const_data(demo_single_table_path):

    const_col_df = pd.read_csv(demo_single_table_path)
    # Convert the columns to float to allow None values
    const_col_df["age"] = const_col_df["age"].astype(float)
    const_col_df["fnlwgt"] = const_col_df["fnlwgt"].astype(float)

    # Set the values to None
    const_col_df["age"].values[:] = 100
    const_col_df["fnlwgt"].values[:] = 1.41421
    const_col_df["workclass"].values[:] = "President"

    yield const_col_df


def test_const_handling_test_df(test_const_data: pd.DataFrame):
    """
    Test the handling of const columns in a DataFrame.
    This function tests the behavior of a DataFrame when it contains const columns.
    It is designed to be used in a testing environment, where the DataFrame is passed as an argument.

    Parameters:
    test_const_data (pd.DataFrame): The DataFrame to test.

    Returns:
    None

    Raises:
    AssertionError: If the DataFrame does not handle const columns as expected.
    """

    metadata = Metadata.from_dataframe(test_const_data)

    # Initialize the ConstValueTransformer.
    const_transformer = ConstValueTransformer()
    # Check if the transformer has not been fitted yet.
    assert const_transformer.fitted is False

    # Fit the transformer with the DataFrame.
    const_transformer.fit(metadata)

    # Check if the transformer has been fitted after the fit operation.
    assert const_transformer.fitted

    # Check the const column
    assert sorted(const_transformer.const_columns) == [
        "age",
        "fnlwgt",
        "workclass",
    ]

    # Transform the DataFrame using the transformer.
    transformed_df = const_transformer.convert(test_const_data)

    assert "age" not in transformed_df.columns
    assert "fnlwgt" not in transformed_df.columns
    assert "workclass" not in transformed_df.columns

    # reverse convert the df
    reverse_converted_df = const_transformer.reverse_convert(transformed_df)

    assert "age" in reverse_converted_df.columns
    assert "fnlwgt" in reverse_converted_df.columns
    assert "workclass" in reverse_converted_df.columns

    assert reverse_converted_df["age"][0] == 100
    assert reverse_converted_df["fnlwgt"][0] == 1.41421
    assert reverse_converted_df["workclass"][0] == "President"

    assert len(reverse_converted_df["age"].unique()) == 1
    assert len(reverse_converted_df["fnlwgt"].unique()) == 1
    assert len(reverse_converted_df["workclass"].unique()) == 1



"""Univariate copulas module."""

from sdgx.models.components.sdv_copulas.univariate.base import (
    BoundedType,
    ParametricType,
    Univariate,
)
from sdgx.models.components.sdv_copulas.univariate.beta import BetaUnivariate
from sdgx.models.components.sdv_copulas.univariate.gamma import GammaUnivariate
from sdgx.models.components.sdv_copulas.univariate.gaussian import GaussianUnivariate
from sdgx.models.components.sdv_copulas.univariate.gaussian_kde import GaussianKDE
from sdgx.models.components.sdv_copulas.univariate.log_laplace import LogLaplace
from sdgx.models.components.sdv_copulas.univariate.student_t import StudentTUnivariate
from sdgx.models.components.sdv_copulas.univariate.truncated_gaussian import (
    TruncatedGaussian,
)
from sdgx.models.components.sdv_copulas.univariate.uniform import UniformUnivariate

__all__ = (
    "BetaUnivariate",
    "GammaUnivariate",
    "GaussianKDE",
    "GaussianUnivariate",
    "TruncatedGaussian",
    "StudentTUnivariate",
    "Univariate",
    "ParametricType",
    "BoundedType",
    "UniformUnivariate",
    "LogLaplace",
)



import pandas as pd
import pytest

from sdgx.data_exporters.csv_exporter import CsvExporter


@pytest.fixture
def csv_exporter():
    yield CsvExporter()


@pytest.fixture
def export_dst(tmp_path):
    filename = tmp_path / "csv-exported.csv"
    filename.unlink(missing_ok=True)
    yield filename
    # filename.unlink(missing_ok=True)


def test_csv_exporter_df(csv_exporter: CsvExporter, export_dst):
    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    csv_exporter.write(export_dst, df)
    pd.testing.assert_frame_equal(df, pd.read_csv(export_dst))


def test_csv_exporter_generator(csv_exporter: CsvExporter, export_dst):
    def generator():
        yield pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        yield pd.DataFrame({"a": [7, 8, 9], "b": [10, 11, 12]})

    df_all = pd.concat(generator(), ignore_index=True)
    csv_exporter.write(export_dst, generator())
    pd.testing.assert_frame_equal(df_all, pd.read_csv(export_dst))


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import faker
import pandas as pd
import pytest
from typing_extensions import Generator

from sdgx.data_connectors.generator_connector import GeneratorConnector
from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.formatters.datetime import DatetimeFormatter

CHUNK_SIZE = 100


@pytest.fixture
def datetime_test_df():
    total_row = 150
    ff = faker.Faker()
    df = pd.DataFrame([ff.date() for i in range(total_row)], columns=["date"])
    return df


def test_datetime_formatter_test_df(datetime_test_df: pd.DataFrame):
    def df_generator():
        yield datetime_test_df

    data_processors = [DatetimeFormatter()]
    dataconnector = GeneratorConnector(df_generator)
    dataloader = DataLoader(dataconnector, chunksize=CHUNK_SIZE)

    metadata = Metadata.from_dataloader(dataloader)
    metadata.datetime_columns = ["date"]
    metadata.discrete_columns = []
    metadata.datetime_format = {"date": "%Y-%m-%d"}

    for d in data_processors:
        d.fit(metadata=metadata, tabular_data=dataloader)

    def chunk_generator() -> Generator[pd.DataFrame, None, None]:
        for chunk in dataloader.iter():
            for d in data_processors:
                chunk = d.convert(chunk)

            assert not chunk.isna().any().any()
            assert not chunk.isnull().any().any()
            yield chunk

    processed_dataloader = DataLoader(
        GeneratorConnector(chunk_generator), identity=dataloader.identity
    )

    df = processed_dataloader.load_all()

    assert not df.isna().any().any()
    assert not df.isnull().any().any()

    reverse_converted_df = df
    for d in data_processors:
        reverse_converted_df = d.reverse_convert(df)

    assert reverse_converted_df.eq(datetime_test_df).all().all()



import re

from sdgx.data_models.inspectors.extension import hookimpl
from sdgx.data_models.inspectors.regex import RegexInspector

chn_last_names = r"[赵|钱|孙|李|周|吴|郑|王|冯|陈|褚|卫|蒋|沈|韩|杨|朱|秦|尤|许|何|吕|施|张|孔|曹|严|华|金|魏|陶|姜|戚|谢|邹|喻|柏|水|窦|章|云|苏|潘|葛|奚|范|彭|郎|鲁|韦|昌|马|苗|凤|花|方|俞|任|袁|柳|酆|鲍|史|唐|费|廉|岑|薛|雷|贺|倪|汤|滕|殷|罗|毕|郝|邬|安|常|乐|于|时|傅|皮|卞|齐|康|伍|余|元|卜|顾|孟|平|黄|和|穆|萧|尹|姚|邵|湛|汪|祁|毛|禹|狄|米|贝|明|臧|计|伏|成|戴|宋|茅|庞|熊|纪|舒|屈|项|祝|董|梁|杜|阮|蓝|闵|席|季|麻|贾|路|娄|江|童|颜|郭|梅|盛|林|刁|徐|邱|骆|夏|蔡|田|樊|胡|凌|霍|虞|万|支|柯|昝|管|卢|莫|经|房|裘|缪|干|解|应|宗|丁|宣|贲|邓|郁|杭|洪|包|诸|左|石|崔|吉|钮|龚|程|嵇|邢|滑|裴|陆|荣|翁|荀|於|惠|甄|麴|家|封|芮|羿|储|靳|汲|邴|糜|松|井|段|富|巫|乌|焦|巴|弓|牧|隗|山|谷|车|侯|宓|蓬|全|郗|班|仰|仲|伊|宫|宁|仇|栾|暴|甘|钭|历|戎|祖|武|符|刘|景|詹|束|龙|叶|幸|司|韶|郜|黎|溥|印|宿|白|怀|蒲|邰|从|鄂|索|咸|籍|卓|蔺|屠|蒙|池|乔|阳|郁|胥|能|苍|闻|莘|翟|谭|贡|劳|逄|姬|申|扶|堵|冉|宰|郦|雍|却|桑|桂|濮|牛|寿|通|边|扈|燕|冀|浦|尚|农|温|别|庄|晏|柴|瞿|充|慕|连|茹|习|宦|艾|鱼|容|向|古|易|慎|戈|廖|庾|终|暨|居|衡|步|都|耿|满|弘|匡|国|文|寇|广|禄|阙|东|欧|沃|利|蔚|越|夔|隆|师|巩|厍|聂|晁|勾|敖|融|冷|訾|辛|阚|那|简|饶|空|曾|毋|沙|乜|养|鞠|须|丰|巢|关|蒯|相|荆|红|游|竺|权|司马|上官|欧阳|夏侯|诸葛|闻人|东方|赫连|皇甫|尉迟|公羊|澹台|公冶宗政|濮阳|淳于|单于|太叔|申屠|公孙|仲孙|轩辕|令狐|钟离|宇文|长孙|慕容|司徒|司空|召|有|舜|岳|黄辰|寸|贰|皇|侨|彤|竭|端|赫|实|甫|集|象|翠|狂|辟|典|良|函|芒|苦|其|京|中|夕|乌孙|完颜|富察|费莫|蹇|称|诺|来|多|繁|戊|朴|回|毓|鉏|税|荤|靖|绪|愈|硕|牢|买|但|巧|枚|撒|泰|秘|亥|绍|以|壬|森|斋|释|奕|姒|朋|求|羽|用|占|真|穰|翦|闾|漆|贵|代|贯|旁|崇|栋|告|休|褒|谏|锐|皋|闳|在|歧|禾|示|是|委|钊|频|嬴|呼|大|威|昂|律|冒|保|系|抄|定|化|莱|校|么|抗|祢|綦|悟|宏|功|庚|务|敏|捷|拱|兆|丑|丙|畅|苟|随|类|卯|俟|友|答|乙|允|甲|留|尾|佼|玄|乘|裔|延|植|环|矫|赛|昔|侍|度|旷|遇|偶|前|由|咎|塞|敛|受|泷|袭|衅|叔|圣|御|夫|仆|镇|藩|邸|府|掌|首|员|焉|戏|可|智|尔|凭|悉|进|笃|厚|仁|肇|资|仍|九|衷|哀|刑|俎|仵|圭|夷|徭|蛮|汗|孛|乾|帖|罕|洛|淦|洋|邶|郸|郯|邗|邛|剑|虢|隋|蒿|茆|菅|苌|树|桐|锁|钟|机|盘|铎|斛|玉|线|针|箕|庹|绳|磨|蒉|瓮|弭|刀|疏|牵|浑|恽|势|世|仝|同|蚁|止|戢|睢|冼|种|涂|肖|己|泣|潜|卷|脱|谬|蹉|赧|浮|顿|说|次|错|念|夙|斯|完|丹|聊|源|吾|寻|展|不|户|闭|才|愚|霜|烟|寒|少|字|桥|板|斐|独|千|诗|嘉|扬|善|揭|祈|析|赤|紫|青|柔|刚|奇|拜|佛|陀|弥|阿|素|长|僧|隐|仙|隽|宇|祭|酒|淡|塔|琦|闪|始|南|天|接|波|碧|速|禚|腾|潮|镜|似|澄|潭|謇|纵|渠|奈|濯|沐|茂|英|兰|檀|藤|枝|检|生|折|登|驹|骑|貊|虎|肥|鹿|雀|野|禽|飞|节|宜|鲜|粟|栗|豆|帛|官|布|衣|藏|宝|钞|银|门|盈|庆|喜|及|普|建|营|巨|望|希|道|载|声|漫|犁|力|贸|勤|革|改|兴|亓|睦|修|信|闽|守|勇|汉|尉|士|旅|五|令|将|旗|军|行|奉|敬|恭|仪|母|堂|丘|义|礼|慈|孝|理|伦|卿|问|永|辉|位|让|尧|依|犹|介|承|市|所|苑|杞|剧|第|零|谌|招|续|达|忻|六|鄞|战|迟|候|宛|励|粘|萨|邝|覃|辜|初|楼|城|区|局|台|原|考|妫|纳|泉|老|清|德|卑|过|麦|曲|竹|百|福|言|第五|佟|爱|年|笪|谯|哈|墨|连|南宫|赏|伯|佴|佘|牟|商|西门|东门|左丘|梁丘|琴|后|况|亢|缑|帅|微生|羊舌|归|呼延|南门|东郭|百里|钦|鄢|汝|法|闫|楚|晋|谷梁|宰父|夹谷|拓跋|壤驷|乐正|漆雕|公西|巫马|端木|颛孙|子车|督|仉|司寇|亓官|三小|鲜于|锺离|盖|逯|库|郏|逢|阴|薄|厉|稽|闾丘|公良|段干|开|光|操|瑞|眭|泥|运|摩|伟|铁|迮]"


class EmailInspector(RegexInspector):
    pattern = r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$"

    data_type_name = "email"

    _inspect_level = 30

    pii = True


class ChinaMainlandIDInspector(RegexInspector):
    pattern = (
        r"^[1-9]\d{5}(18|19|20)\d{2}((0[1-9])|(1[0-2]))(([0-2][1-9])|10|20|30|31)\d{3}[0-9Xx]$"
    )

    data_type_name = "china_mainland_id"

    _inspect_level = 30

    pii = True


class ChinaMainlandMobilePhoneInspector(RegexInspector):
    pattern = r"^1[3-9]\d{9}$"

    data_type_name = "china_mainland_mobile_phone"

    _inspect_level = 30

    pii = True


# 邮编
class ChinaMainlandPostCode(RegexInspector):
    pattern = r"^[0-9]{6}$"

    _match_percentage = 0.95
    """
    Since zip codes and six-digit integers are the same, here we increase match_percentage to prevent some pure integer columns from being recognized.
    """

    data_type_name = "china_mainland_postcode"

    _inspect_level = 20

    pii = False


# 统一社会信用代码
class ChinaMainlandUnifiedSocialCreditCode(RegexInspector):
    pattern = r"^[0-9A-HJ-NPQRTUWXY]{2}\d{6}[0-9A-HJ-NPQRTUWXY]{10}$"

    data_type_name = "unified_social_credit_code"

    _inspect_level = 30

    pii = True

    pattern_ID = (
        r"^[1-9]\d{5}(18|19|20)\d{2}((0[1-9])|(1[0-2]))(([0-2][1-9])|10|20|30|31)\d{3}[0-9Xx]$"
    )

    p_id = re.compile(pattern_ID)

    def domain_verification(self, each_sample):
        if re.match(self.p_id, each_sample):
            return False
        return True


class ChinaMainlandAddressInspector(RegexInspector):

    # This regular expression does not take effect and is only for reference by developers.
    # pattern = r"^[\u4e00-\u9fa5]{2,}(省|自治区|特别行政区|市)|[\u4e00-\u9fa5]{2,}(市|区|县|自治州|自治县|县级市|地区|盟|林区)?|[\u4e00-\u9fa5]{0,}(街道|镇|乡)?|[\u4e00-\u9fa5]{0,}(路|街|巷|弄)?|[\u4e00-\u9fa5]{0,}(号|弄)?$"

    pattern = r"^[\u4e00-\u9fa5]{2,}(省|自治区|特别行政区|市|县|村|弄|乡|路|街)"

    pii = True

    data_type_name = "china_mainland_address"

    _inspect_level = 30

    address_min_length = 8

    address_max_length = 30

    def domain_verification(self, each_sample):
        # CHN address should be between 8 - 30 characters
        if len(each_sample) < self.address_min_length:
            return False
        if len(each_sample) > self.address_max_length:
            return False
        # notice to distinguishing from the company name
        if each_sample.endswith("公司"):
            return False
        return True


# 中文姓名
class ChineseNameInspector(RegexInspector):

    pattern = chn_last_names[:600] + r"][\u4e00-\u9fa5]{1,3}$"

    data_type_name = "chinese_name"

    _inspect_level = 40

    pii = True

    def domain_verification(self, each_sample):
        def has_symbols(s):
            return bool(re.search(r"[^\w\s]", s))

        def has_english(s):
            return bool(re.search(r"[a-zA-Z]", s))

        def has_number(s):
            for char in s:
                if char.isdigit():
                    return True
            return False

        if has_number(each_sample):
            return False
        if has_english(each_sample):
            return False
        if has_symbols(each_sample):
            return False

        return True


# English Name
class EnglishNameInspector(RegexInspector):
    pattern = r"^([a-zA-Z]{2,}\s[a-zA-Z]{1,}'?-?[a-zA-Z]{2,}\s?([a-zA-Z]{1,})?)"

    data_type_name = "english_name"

    _inspect_level = 40

    pii = True

    name_min_length = 5
    """
    The min length of the name.

    GPT-4: The shortest full name in English could be something like "Ed Li" or "Al Lu", with just four characters including a space.
    """

    name_max_length = 70
    """
    The max length of the name.

    UK Government Data Standards Catalogue suggests 35 characters for each of Given Name and Family Name, or 70 characters for a single field to hold the Full Name.
    """

    def domain_verification(self, each_sample):
        def has_number(s):
            for char in s:
                if char.isdigit():
                    return True
            return False

        # English name should be between 5 - 70 characters
        if len(each_sample) > self.name_max_length:
            return False
        if len(each_sample) < self.name_min_length:
            return False
        # usually a name should not contains number
        if has_number(each_sample):
            return False
        return True


# 公司名
class ChineseCompanyNameInspector(RegexInspector):
    pattern = r".*?公司.*?"

    _match_percentage = 0.7

    data_type_name = "chinese_company_name"

    _inspect_level = 40

    pii = False


@hookimpl
def register(manager):
    manager.register("EmailInspector", EmailInspector)

    manager.register("ChinaMainlandIDInspector", ChinaMainlandIDInspector)

    manager.register("ChinaMainlandMobilePhoneInspector", ChinaMainlandMobilePhoneInspector)

    manager.register("ChinaMainlandPostCode", ChinaMainlandPostCode)

    manager.register("ChinaMainlandUnifiedSocialCreditCode", ChinaMainlandUnifiedSocialCreditCode)

    manager.register("ChinaMainlandAddressInspector", ChinaMainlandAddressInspector)

    manager.register("ChineseNameInspector", ChineseNameInspector)

    manager.register("EnglishNameInspector", EnglishNameInspector)

    manager.register("ChineseCompanyNameInspector", ChineseCompanyNameInspector)



from __future__ import annotations

from typing import Any

import pandas as pd

from sdgx.data_models.inspectors.base import Inspector
from sdgx.data_models.inspectors.extension import hookimpl


class EmptyInspector(Inspector):
    """
    The EmptyInspector class is designed to identify columns in a DataFrame that have a high rate of missing values.

    Columns taged empty will be removed during the training process and reinserted into their original positions after the model sampling process is complete.

    Attributes:
        empty_rate_threshold (float): The threshold for the rate of missing values above which a column is considered empty, default = 0.9.
        empty_columns (set[str]): A set of column names that have missing values above the threshold.

    Methods:
        __init__(self, *args, **kwargs): Initializes the EmptyInspector instance, optionally setting the empty_rate_threshold.
        fit(self, raw_data: pd.DataFrame, *args, **kwargs): Fits the inspector to the raw data, identifying columns with missing values above the threshold.
        inspect(self, *args, **kwargs) -> dict[str, Any]: Returns a dictionary containing the list of columns identified as empty.
    """

    empty_rate_threshold = 0.9
    """
    float: The threshold for the rate of missing values above which a column is considered empty.
    Default is 0.9, meaning if a column has more than 90% of its values missing, it will be considered empty.
    """

    empty_columns: set[str] = set()
    """
    set[str]: A set of column names that have missing values above the empty_rate_threshold.
    These columns are identified as empty and will be handled accordingly during the data processing.
    """

    _inspect_level = 90
    """
    int: The inspection level for the EmptyInspector, set to a quite high value (90) to prioritize the identification and handling of empty columns.
    This high value is chosen because empty columns contain no information and should not be considered for any other type of inspection or processing.
    They are typically removed during model training as they cannot be understood by many models and may cause errors.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if "empty_rate_threshold" in kwargs:
            self.empty_rate_threshold = kwargs["empty_rate_threshold"]

    def fit(self, raw_data: pd.DataFrame, *args, **kwargs):
        """Fit the inspector.

        Gets the list of empty columns from the raw data.

        Args:
            raw_data (pd.DataFrame): Raw data
        """
        # Calculate the empty rate for each column
        empty_rate = raw_data.isnull().mean()

        # Identify columns where the empty rate exceeds the threshold
        self.empty_columns = set(empty_rate[empty_rate >= self.empty_rate_threshold].index)

        self.ready = True

    def inspect(self, *args, **kwargs) -> dict[str, Any]:
        """Inspect raw data and generate metadata."""

        return {"empty_columns": list(self.empty_columns)}


@hookimpl
def register(manager):
    manager.register("EmptyInspector", EmptyInspector)



"""This module contains a base class for bivariate copulas."""

import json
import warnings
from enum import Enum

import numpy as np
from scipy import stats
from scipy.optimize import brentq

from sdgx.models.components.sdv_copulas import (
    EPSILON,
    NotFittedError,
    random_state,
    validate_random_state,
)
from sdgx.models.components.sdv_copulas.bivariate.utils import split_matrix


class CopulaTypes(Enum):
    """Available copula families."""

    CLAYTON = 0
    FRANK = 1
    GUMBEL = 2
    INDEPENDENCE = 3


class Bivariate(object):
    """Base class for bivariate copulas.

    This class allows to instantiate all its subclasses and serves as a unique entry point for
    the bivariate copulas classes.

    >>> Bivariate(copula_type=CopulaTypes.FRANK).__class__
    copulas.bivariate.frank.Frank

    >>> Bivariate(copula_type='frank').__class__
    copulas.bivariate.frank.Frank


    Args:
        copula_type (Union[CopulaType, str]): Subtype of the copula.
        random_state (Union[int, np.random.RandomState, None]): Seed or RandomState
            for the random generator.

    Attributes:
        copula_type(CopulaTypes): Family of the copula a subclass belongs to.
        _subclasses(list[type]): List of declared subclasses.
        theta_interval(list[float]): Interval of valid thetas for the given copula family.
        invalid_thetas(list[float]): Values that, even though they belong to
            :attr:`theta_interval`, shouldn't be considered valid.
        tau (float): Kendall's tau for the data given at :meth:`fit`.
        theta(float): Parameter for the copula.

    """

    copula_type = None
    _subclasses = []
    theta_interval = []
    invalid_thetas = []
    theta = None
    tau = None

    @classmethod
    def _get_subclasses(cls):
        """Find recursively subclasses for the current class object.

        Returns:
            list[Bivariate]: List of subclass objects.

        """
        subclasses = []
        for subclass in cls.__subclasses__():
            subclasses.append(subclass)
            subclasses.extend(subclass._get_subclasses())

        return subclasses

    @classmethod
    def subclasses(cls):
        """Return a list of subclasses for the current class object.

        Returns:
            list[Bivariate]: Subclasses for given class.

        """
        if not cls._subclasses:
            cls._subclasses = cls._get_subclasses()

        return cls._subclasses

    def __new__(cls, *args, **kwargs):
        """Create and return a new object.

        Returns:
            Bivariate: New object.
        """
        copula_type = kwargs.get("copula_type", None)
        if copula_type is None:
            return super(Bivariate, cls).__new__(cls)

        if not isinstance(copula_type, CopulaTypes):
            if isinstance(copula_type, str) and copula_type.upper() in CopulaTypes.__members__:
                copula_type = CopulaTypes[copula_type.upper()]
            else:
                raise ValueError(f"Invalid copula type {copula_type}")

        for subclass in cls.subclasses():
            if subclass.copula_type is copula_type:
                return super(Bivariate, cls).__new__(subclass)

    def __init__(self, copula_type=None, random_state=None):
        """Initialize Bivariate object.

        Args:
            copula_type (CopulaType or str): Subtype of the copula.
            random_state (int, np.random.RandomState, or None): Seed or RandomState
                for the random generator.
        """
        self.random_state = validate_random_state(random_state)

    def check_theta(self):
        """Validate the computed theta against the copula specification.

        This method is used to assert the computed theta is in the valid range for the copula.

        Raises:
            ValueError: If theta is not in :attr:`theta_interval` or is in :attr:`invalid_thetas`,

        """
        lower, upper = self.theta_interval
        if (not lower <= self.theta <= upper) or (self.theta in self.invalid_thetas):
            message = "The computed theta value {} is out of limits for the given {} copula."
            raise ValueError(message.format(self.theta, self.copula_type.name))

    def check_fit(self):
        """Assert that the model is fit and the computed `theta` is valid.

        Raises:
            NotFittedError: if the model is  not fitted.
            ValueError: if the computed theta is invalid.

        """
        if not self.theta:
            raise NotFittedError("This model is not fitted.")

        self.check_theta()

    def check_marginal(self, u):
        """Check that the marginals are uniformly distributed.

        Args:
            u(np.ndarray): Array of datapoints with shape (n,).

        Raises:
            ValueError: If the data does not appear uniformly distributed.
        """
        if min(u) < 0.0 or max(u) > 1.0:
            raise ValueError("Marginal value out of bounds.")

        emperical_cdf = np.sort(u)
        uniform_cdf = np.linspace(0.0, 1.0, num=len(u))
        ks_statistic = max(np.abs(emperical_cdf - uniform_cdf))
        if ks_statistic > 1.627 / np.sqrt(len(u)):
            # KS test with significance level 0.01
            warnings.warn("Data does not appear to be uniform.", category=RuntimeWarning)

    def _compute_theta(self):
        """Compute theta, validate it and assign it to self."""
        self.theta = self.compute_theta()
        self.check_theta()

    def fit(self, X):
        """Fit a model to the data updating the parameters.

        Args:
            X(np.ndarray): Array of datapoints with shape (n,2).

        Return:
            None
        """
        U, V = split_matrix(X)
        self.check_marginal(U)
        self.check_marginal(V)
        self.tau = stats.kendalltau(U, V)[0]
        if np.isnan(self.tau):
            if len(np.unique(U)) == 1 or len(np.unique(V)) == 1:
                raise ValueError("Constant column.")
            raise ValueError("Unable to compute tau.")
        self._compute_theta()

    def to_dict(self):
        """Return a `dict` with the parameters to replicate this object.

        Returns:
            dict: Parameters of the copula.

        """
        return {"copula_type": self.copula_type.name, "theta": self.theta, "tau": self.tau}

    @classmethod
    def from_dict(cls, copula_dict):
        """Create a new instance from the given parameters.

        Args:
            copula_dict: `dict` with the parameters to replicate the copula.
              Like the output of `Bivariate.to_dict`

        Returns:
            Bivariate: Instance of the copula defined on the parameters.

        """
        instance = cls(copula_type=copula_dict["copula_type"])
        instance.theta = copula_dict["theta"]
        instance.tau = copula_dict["tau"]
        return instance

    def infer(self, X):
        """Take in subset of values and predicts the rest."""
        raise NotImplementedError

    def generator(self, t):
        r"""Compute the generator function for Archimedian copulas.

        The generator is a function
        :math:`\psi: [0,1]\times\Theta \rightarrow [0, \infty)`  # noqa: JS101

        that given an Archimedian copula fulfills:
        .. math:: C(u,v) = \psi^{-1}(\psi(u) + \psi(v))


        In a more generic way:

        .. math:: C(u_1, u_2, ..., u_n;\theta) = \psi^-1(\sum_0^n{\psi(u_i;\theta)}; \theta)

        """
        raise NotImplementedError

    def probability_density(self, X):
        r"""Compute probability density function for given copula family.

        The probability density(pdf) for a given copula is defined as:

        .. math:: c(U,V) = \frac{\partial^2 C(u,v)}{\partial v \partial u}

        Args:
            X(np.ndarray): Shape (n, 2).Datapoints to compute pdf.

        Returns:
            np.array: Probability density for the input values.

        """
        raise NotImplementedError

    def log_probability_density(self, X):
        """Return log probability density of model.

        The log probability should be overridden with numerically stable
        variants whenever possible.

        Arguments:
            X: `np.ndarray` of shape (n, 1).

        Returns:
            np.ndarray

        """
        return np.log(self.probability_density(X))

    def pdf(self, X):
        """Shortcut to :meth:`probability_density`."""
        return self.probability_density(X)

    def cumulative_distribution(self, X):
        """Compute the cumulative distribution function for the copula, :math:`C(u, v)`.

        Args:
            X(np.ndarray):

        Returns:
            numpy.array: cumulative probability

        """
        raise NotImplementedError

    def cdf(self, X):
        """Shortcut to :meth:`cumulative_distribution`."""
        return self.cumulative_distribution(X)

    def percent_point(self, y, V):
        """Compute the inverse of conditional cumulative distribution :math:`C(u|v)^{-1}`.

        Args:
            y: `np.ndarray` value of :math:`C(u|v)`.
            v: `np.ndarray` given value of v.
        """
        self.check_fit()
        result = []
        for _y, _v in zip(y, V):

            def f(u):
                return self.partial_derivative_scalar(u, _v) - _y

            minimum = brentq(f, EPSILON, 1.0)
            if isinstance(minimum, np.ndarray):
                minimum = minimum[0]

            result.append(minimum)

        return np.array(result)

    def ppf(self, y, V):
        """Shortcut to :meth:`percent_point`."""
        return self.percent_point(y, V)

    def partial_derivative(self, X):
        r"""Compute partial derivative of cumulative distribution.

        The partial derivative of the copula(CDF) is the conditional CDF.

         .. math:: F(v|u) = \frac{\partial C(u,v)}{\partial u}

        The base class provides a finite difference approximation of the
        partial derivative of the CDF with respect to u.

        Args:
            X(np.ndarray)
            y(float)

        Returns:
            np.ndarray

        """
        delta = -2 * (X[:, 1] > 0.5) + 1
        delta = 0.0001 * delta
        X_prime = X.copy()
        X_prime[:, 1] += delta
        f = self.cumulative_distribution(X)
        f_prime = self.cumulative_distribution(X_prime)
        return (f_prime - f) / delta

    def partial_derivative_scalar(self, U, V):
        """Compute partial derivative :math:`C(u|v)` of cumulative density of single values."""
        self.check_fit()

        X = np.column_stack((U, V))
        return self.partial_derivative(X)

    def set_random_state(self, random_state):
        """Set the random state.

        Args:
            random_state (int, np.random.RandomState, or None): Seed or RandomState
                for the random generator.
        """
        self.random_state = validate_random_state(random_state)

    @random_state
    def sample(self, n_samples):
        """Generate specified `n_samples` of new data from model.

        The sampled are generated using the inverse transform method `v~U[0,1],v~C^-1(u|v)`

        Args:
            n_samples (int): amount of samples to create.

        Returns:
            np.ndarray: Array of length `n_samples` with generated data from the model.

        """
        if self.tau > 1 or self.tau < -1:
            raise ValueError("The range for correlation measure is [-1,1].")

        v = np.random.uniform(0, 1, n_samples)
        c = np.random.uniform(0, 1, n_samples)

        u = self.percent_point(c, v)
        return np.column_stack((u, v))

    def compute_theta(self):
        """Compute theta parameter using Kendall's tau."""
        raise NotImplementedError

    @classmethod
    def select_copula(cls, X):
        r"""Select best copula function based on likelihood.

        Given out candidate copulas the procedure proposed for selecting the one
        that best fit to a dataset of pairs :math:`\{(u_j, v_j )\}, j=1,2,...n` , is as follows:

        1. Estimate the most likely parameter :math:`\theta` of each copula candidate for the given
           dataset.

        2. Construct :math:`R(z|\theta)`. Calculate the area under the tail for each of the copula
           candidates.

        3. Compare the areas: :math:`a_u` achieved using empirical copula against the ones
           achieved for the copula candidates. Score the outcome of the comparison from 3 (best)
           down to 1 (worst).

        4. Proceed as in steps 2- 3 with the lower tail and function :math:`L`.

        5. Finally the sum of empirical upper and lower tail functions is compared against
           :math:`R + L`. Scores of the three comparisons are summed and the candidate with the
           highest value is selected.

        Args:
            X(np.ndarray): Matrix of shape (n,2).

        Returns:
            copula: Best copula that fits for it.

        """
        from sdgx.models.components.sdv_copulas.bivariate import select_copula  # noqa

        warnings.warn(
            "`Bivariate.select_copula` has been deprecated and will be removed in a later "
            "release. Please use `copulas.bivariate.select_copula` instead",
            DeprecationWarning,
        )
        return select_copula(X)

    def save(self, filename):
        """Save the internal state of a copula in the specified filename.

        Args:
            filename(str): Path to save.

        Returns:
            None

        """
        content = self.to_dict()
        with open(filename, "w") as f:
            json.dump(content, f)

    @classmethod
    def load(cls, copula_path):
        """Create a new instance from a file.

        Args:
            copula_path(str): Path to file with the serialized copula.

        Returns:
            Bivariate: Instance with the parameters stored in the file.

        """
        with open(copula_path) as f:
            copula_dict = json.load(f)

        return cls.from_dict(copula_dict)



from __future__ import annotations

from typing import Any

from sdgx import data_exporters
from sdgx.data_exporters import extension
from sdgx.data_exporters.base import DataExporter
from sdgx.data_exporters.extension import project_name as PROJECT_NAME
from sdgx.manager import Manager


class DataExporterManager(Manager):
    register_type = DataExporter
    project_name = PROJECT_NAME
    hookspecs_model = extension

    @property
    def registed_exporters(self):
        return self.registed_cls

    def load_all_local_model(self):
        self._load_dir(data_exporters)

    def init_exporter(self, exporter_name, **kwargs: dict[str, Any]) -> DataExporter:
        return self.init(exporter_name, **kwargs)



import logging
import warnings
from copy import deepcopy

import numpy as np
import pandas as pd
import scipy

import sdgx.models.components.sdv_copulas as copulas
from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.exceptions import NonParametricError
from sdgx.models.components.optimize.sdv_copulas.data_transformer import (
    StatisticDataTransformer,
)
from sdgx.models.components.sdv_copulas import multivariate
from sdgx.models.components.utils import (
    flatten_dict,
    unflatten_dict,
    validate_numerical_distributions,
)
from sdgx.models.statistics.single_table.base import StatisticSynthesizerModel

LOGGER = logging.getLogger(__name__)


class GaussianCopulaSynthesizerModel(StatisticSynthesizerModel):
    """Model wrapping ``copulas.multivariate.GaussianMultivariate`` copula.

    Args:
        metadata (sdgx.data_models.metadata.Metadata):
            Metadata of the input table.
        enforce_min_max_values (bool):
            Specify whether or not to clip the data returned by ``reverse_transform`` of
            the numerical transformer, ``FloatFormatter``, to the min and max values seen
            during ``fit``. Defaults to ``True``.
        enforce_rounding (bool):
            Define rounding scheme for ``numerical`` columns. If ``True``, the data returned
            by ``reverse_transform`` will be rounded as in the original data. Defaults to ``True``.
        locales (list or str):
            The default locale(s) to use for AnonymizedFaker transformers. Defaults to ``None``.
        numerical_distributions (dict):
            Dictionary that maps field names from the table that is being modeled with
            the distribution that needs to be used. The distributions can be passed as either
            a ``copulas.univariate`` instance or as one of the following values:

                * ``norm``: Use a norm distribution.
                * ``beta``: Use a Beta distribution.
                * ``truncnorm``: Use a truncnorm distribution.
                * ``uniform``: Use a uniform distribution.
                * ``gamma``: Use a Gamma distribution.
                * ``gaussian_kde``: Use a GaussianKDE distribution. This model is non-parametric,
                  so using this will make ``get_parameters`` unusable.

        default_distribution (str):
            Copulas univariate distribution to use by default. Valid options are:

                * ``norm``: Use a norm distribution.
                * ``beta``: Use a Beta distribution.
                * ``truncnorm``: Use a Truncated Gaussian distribution.
                * ``uniform``: Use a uniform distribution.
                * ``gamma``: Use a Gamma distribution.
                * ``gaussian_kde``: Use a GaussianKDE distribution. This model is non-parametric,
                  so using this will make ``get_parameters`` unusable.
             Defaults to ``beta``.
    """

    _DISTRIBUTIONS = {
        "norm": copulas.univariate.GaussianUnivariate,
        "beta": copulas.univariate.BetaUnivariate,
        "truncnorm": copulas.univariate.TruncatedGaussian,
        "gamma": copulas.univariate.GammaUnivariate,
        "uniform": copulas.univariate.UniformUnivariate,
        "gaussian_kde": copulas.univariate.GaussianKDE,
    }

    _model = None

    @classmethod
    def get_distribution_class(cls, distribution):
        """Return the corresponding distribution class from ``copulas.univariate``.

        Args:
            distribution (str):
                A string representing a copulas univariate distribution.

        Returns:
            copulas.univariate:
                A copulas univariate class that corresponds to the distribution.
        """
        if not isinstance(distribution, str) or distribution not in cls._DISTRIBUTIONS:
            error_message = f"Invalid distribution specification '{distribution}'."
            raise ValueError(error_message)

        return cls._DISTRIBUTIONS[distribution]

    def __init__(
        self,
        metadata: Metadata = None,
        enforce_min_max_values=True,
        enforce_rounding=True,
        locales=None,
        numerical_distributions=None,
        default_distribution=None,
    ):
        self.metadata = metadata
        self.enforce_min_max_values = (enforce_min_max_values,)
        self.enforce_rounding = (enforce_rounding,)
        self.locales = (locales,)

        if isinstance(self.metadata, Metadata):
            self.discrete_cols = self.metadata.discrete_columns
        else:
            self.discrete_cols = None

        validate_numerical_distributions(numerical_distributions, self.metadata)

        self.numerical_distributions = numerical_distributions or {}
        self.default_distribution = default_distribution or "beta"

        self._default_distribution = self.get_distribution_class(self.default_distribution)
        self._numerical_distributions = {
            field: self.get_distribution_class(distribution)
            for field, distribution in self.numerical_distributions.items()
        }

        self._num_rows = None
        self._transformer = None

    def fit(self, metadata: Metadata, dataloader: DataLoader, *args, **kwargs):

        # extract pd.DataFrame processed_data from dataloader
        processed_data: pd.DataFrame = dataloader.load_all()

        # get discrete_cols from metadata
        self.discrete_cols = list(metadata.get("discrete_columns"))

        self.metadata = metadata

        # load the original transformer
        self._transformer = StatisticDataTransformer()

        # self._transformer.fit(processed_data, self.metadata[0])
        self._transformer.fit(processed_data, self.discrete_cols)

        processed_data = pd.DataFrame(self._transformer.transform(processed_data))

        """
        log_numerical_distributions_error(
            self.numerical_distributions, processed_data.columns, LOGGER
        )
        """

        self._num_rows = len(processed_data)

        numerical_distributions = deepcopy(self._numerical_distributions)
        for column in processed_data.columns:
            if column not in numerical_distributions:
                numerical_distributions[column] = self._numerical_distributions.get(
                    column, self._default_distribution
                )

        self._model = multivariate.GaussianMultivariate(distribution=numerical_distributions)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", module="scipy")
            self._model.fit(processed_data)

    def sample(self, num_rows, conditions=None):
        """Sample the indicated number of rows from the model.

        Args:
            num_rows (int):
                Amount of rows to sample.
            conditions (dict):
                If specified, this dictionary maps column names to the column
                value. Then, this method generates ``num_rows`` samples, all of
                which are conditioned on the given variables.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        return self._transformer.inverse_transform(
            self._model.sample(num_rows, conditions=conditions).to_numpy()
        )

    def _get_valid_columns_from_metadata(self, columns):
        valid_columns = []
        for column in columns:
            # for valid_column in self.metadata.columns:
            for valid_column in self.metadata.column_list:
                if column.startswith(valid_column):
                    valid_columns.append(column)
                    break

        return valid_columns

    def get_learned_distributions(self):
        """Get the marginal distributions used by the ``GaussianCopula``.

        Return a dictionary mapping the column names with the distribution name and the learned
        parameters for those.

        Returns:
            dict:
                Dictionary containing the distributions used or detected for each column and the
                learned parameters for those.
        """
        if not self._fitted:
            raise ValueError(
                "Distributions have not been learned yet. Please fit your model first using 'fit'."
            )

        parameters = self._model.to_dict()
        columns = parameters["columns"]
        univariates = deepcopy(parameters["univariates"])
        learned_distributions = {}
        valid_columns = self._get_valid_columns_from_metadata(columns)
        for column, learned_params in zip(columns, univariates):
            if column in valid_columns:
                distribution = self.numerical_distributions.get(column, self.default_distribution)
                learned_params.pop("type")
                learned_distributions[column] = {
                    "distribution": distribution,
                    "learned_parameters": learned_params,
                }

        return learned_distributions

    def _get_parameters(self):
        """Get copula model parameters.

        Compute model ``correlation`` and ``distribution.std``
        before it returns the flatten dict.

        Returns:
            dict:
                Copula parameters.

        Raises:
            NonParametricError:
                If a non-parametric distribution has been used.
        """
        for univariate in self._model.univariates:
            univariate_type = type(univariate)
            if univariate_type is copulas.univariate.Univariate:
                univariate = univariate._instance

            if univariate.PARAMETRIC == copulas.univariate.ParametricType.NON_PARAMETRIC:
                raise NonParametricError("This GaussianCopula uses non parametric distributions")

        params = self._model.to_dict()

        correlation = []
        for index, row in enumerate(params["correlation"][1:]):
            correlation.append(row[: index + 1])

        params["correlation"] = correlation
        params["univariates"] = dict(zip(params.pop("columns"), params["univariates"]))
        params["num_rows"] = self._num_rows

        return flatten_dict(params)

    @staticmethod
    def _get_nearest_correlation_matrix(matrix):
        """Find the nearest correlation matrix.

        If the given matrix is not Positive Semi-definite, which means
        that any of its eigenvalues is negative, find the nearest PSD matrix
        by setting the negative eigenvalues to 0 and rebuilding the matrix
        from the same eigenvectors and the modified eigenvalues.

        After this, the matrix will be PSD but may not have 1s in the diagonal,
        so the diagonal is replaced by 1s and then the PSD condition of the
        matrix is validated again, repeating the process until the built matrix
        contains 1s in all the diagonal and is PSD.

        After 10 iterations, the last step is skipped and the current PSD matrix
        is returned even if it does not have all 1s in the diagonal.

        Insipired by: https://stackoverflow.com/a/63131250
        """
        eigenvalues, eigenvectors = scipy.linalg.eigh(matrix)
        negative = eigenvalues < 0
        identity = np.identity(len(matrix))

        iterations = 0
        while np.any(negative):
            eigenvalues[negative] = 0
            matrix = eigenvectors.dot(np.diag(eigenvalues)).dot(eigenvectors.T)
            if iterations >= 10:
                break

            matrix = matrix - matrix * identity + identity

            max_value = np.abs(np.abs(matrix).max())
            if max_value > 1:
                matrix /= max_value

            eigenvalues, eigenvectors = scipy.linalg.eigh(matrix)
            negative = eigenvalues < 0
            iterations += 1

        return matrix

    @classmethod
    def _rebuild_correlation_matrix(cls, triangular_correlation):
        """Rebuild a valid correlation matrix from its lower half triangle.

        The input of this function is a list of lists of floats of size 1, 2, 3...n-1:

           [[c_{2,1}], [c_{3,1}, c_{3,2}], ..., [c_{n,1},...,c_{n,n-1}]]

        Corresponding to the values from the lower half of the original correlation matrix,
        **excluding** the diagonal.

        The output is the complete correlation matrix reconstructed using the given values
        and scaled to the :math:`[-1, 1]` range if necessary.

        Args:
            triangle_correlation (list[list[float]]):
                A list that contains lists of floats of size 1, 2, 3... up to ``n-1``,
                where ``n`` is the size of the target correlation matrix.

        Returns:
            numpy.ndarray:
                rebuilt correlation matrix.
        """
        zero = [0.0]
        size = len(triangular_correlation) + 1
        left = np.zeros((size, size))
        right = np.zeros((size, size))
        for idx, values in enumerate(triangular_correlation):
            values = values + zero * (size - idx - 1)
            left[idx + 1, :] = values
            right[:, idx + 1] = values

        correlation = left + right
        max_value = np.abs(correlation).max()
        if max_value > 1:
            correlation /= max_value

        correlation += np.identity(size)

        return cls._get_nearest_correlation_matrix(correlation).tolist()

    def _rebuild_gaussian_copula(self, model_parameters):
        """Rebuild the model params to recreate a Gaussian Multivariate instance.

        Args:
            model_parameters (dict):
                Sampled and reestructured model parameters.

        Returns:
            dict:
                Model parameters ready to recreate the model.
        """
        columns = []
        univariates = []
        for column, univariate in model_parameters["univariates"].items():
            columns.append(column)
            univariate["type"] = self.get_distribution_class(
                self._numerical_distributions.get(column, self.default_distribution)
            )
            if "scale" in univariate:
                univariate["scale"] = max(0, univariate["scale"])

            univariates.append(univariate)

        model_parameters["univariates"] = univariates
        model_parameters["columns"] = columns

        correlation = model_parameters.get("correlation")
        if correlation:
            model_parameters["correlation"] = self._rebuild_correlation_matrix(correlation)
        else:
            model_parameters["correlation"] = [[1.0]]

        return model_parameters

    def _get_likelihood(self, table_rows):
        return self._model.probability_density(table_rows)

    def _set_parameters(self, parameters):
        """Set copula model parameters.

        Args:
            dict:
                Copula flatten parameters.
        """
        parameters = unflatten_dict(parameters)
        if "num_rows" in parameters:
            num_rows = parameters.pop("num_rows")
            self._num_rows = 0 if pd.isna(num_rows) else max(0, int(round(num_rows)))

        if parameters:
            parameters = self._rebuild_gaussian_copula(parameters)
            self._model = multivariate.GaussianMultivariate.from_dict(parameters)



from __future__ import annotations

from typing import Any

from sdgx import data_processors
from sdgx.data_processors import extension
from sdgx.data_processors.base import DataProcessor
from sdgx.data_processors.extension import project_name as PROJECT_NAME
from sdgx.manager import Manager


class DataProcessorManager(Manager):
    """
    This is a plugin management class for data processing components.

    Properties:
        - register_type: Specifies the type of data processors to register.
        - project_name: Stores the project name from the extension module.
        - hookspecs_model: Stores the hook specifications model from the extension module.
        - preset_default_processors: Stores a list of default processor names in lowercase.
        - registed_data_processors: Property that returns the registered data processors.
        - registed_default_processor_list: Property that returns the registered default data processors.

    Methods:
        - load_all_local_model: Loads all local models for formatters, generators, samplers, and transformers.
        - init_data_processor: Initializes a data processor with the given name and keyword arguments.
        - init_all_processors: Initializes all registered data processors with optional keyword arguments.
        - init_default_processors: Initializes default processors that are both registered and preset.

    """

    register_type = DataProcessor
    """
    Specifies the type of data processors to register."""

    project_name = PROJECT_NAME
    """
    Stores the project name from the extension module.
    """

    hookspecs_model = extension
    """
    The hook specifications model from the extension module.
    """

    preset_defalut_processors = [
        p.lower()
        for p in [
            "SpecificCombinationTransformer",
            "FixedCombinationTransformer",
            "NonValueTransformer",
            "OutlierTransformer",
            "EmailGenerator",
            "ChnPiiGenerator",
            "IntValueFormatter",
            "DatetimeFormatter",
        ]
    ] + [
        "ConstValueTransformer".lower(),
        "PositiveNegativeFilter".lower(),
        "EmptyTransformer".lower(),
        "ColumnOrderTransformer".lower(),
    ]
    """
    preset_defalut_processors list stores the lowercase names of the transformers loaded by default. When using the synthesizer, they will be loaded by default to facilitate user operations.

    Keep ColumnOrderTransformer always at the last one.
    """

    @property
    def registed_data_processors(self):
        """
        This property returns all registered data processors
        """
        return self.registed_cls

    @property
    def registed_default_processor_list(self):
        """
        This property returns all registered default data processors
        """
        registed_processor_list = self.registed_data_processors.keys()

        # we donot use the code next line to calculate processor intersection, to ensure the oder of processors not changed.
        # target_processors = list(set(registed_processor_list).intersection(self.preset_defalut_processors))

        # calculate the target_processors
        default_processors = []
        for each_processor in self.preset_defalut_processors:
            if each_processor in registed_processor_list:
                default_processors.append(each_processor)

        return default_processors

    def load_all_local_model(self):
        """
        loads all local models
        """
        self._load_dir(data_processors.formatters)
        self._load_dir(data_processors.generators)
        self._load_dir(data_processors.samplers)
        self._load_dir(data_processors.transformers)
        self._load_dir(data_processors.filter)

    def init_data_processor(self, processor_name, **kwargs: dict[str, Any]) -> DataProcessor:
        """
        Initializes a data processor with the given name and parameters
        """
        return self.init(processor_name, **kwargs)

    def init_all_processors(self, **kwargs: Any) -> list[DataProcessor]:
        """
        Initializes all registered data processors
        """
        return [
            self.init(processor_name, **kwargs)
            for processor_name in self.registed_data_processors.keys()
        ]

    def init_default_processors(self, **kwargs: Any) -> list[DataProcessor]:
        """
        Initializes all default data processors
        """

        return [
            self.init(processor_name, **kwargs)
            for processor_name in self.registed_default_processor_list
        ]



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from sdgx.data_connectors.generator_connector import GeneratorConnector


def generator_caller():
    yield pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    yield pd.DataFrame({"a": [7, 8, 9], "b": [10, 11, 12]})
    yield pd.DataFrame({"a": [13, 14, 15], "b": [16, 17, 18]})


def test_generator_connector():
    c = GeneratorConnector(generator_caller)
    assert c._columns() == ["a", "b"]
    assert_frame_equal(c._read(), pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}))
    assert_frame_equal(c._read(offset=1), pd.DataFrame({"a": [7, 8, 9], "b": [10, 11, 12]}))
    assert_frame_equal(c._read(offset=2), pd.DataFrame({"a": [13, 14, 15], "b": [16, 17, 18]}))
    assert c._read(offset=3) is None

    # Reset
    assert_frame_equal(c._read(offset=0), pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}))
    assert_frame_equal(c._read(offset=1000), pd.DataFrame({"a": [7, 8, 9], "b": [10, 11, 12]}))
    assert_frame_equal(c._read(offset=3), pd.DataFrame({"a": [13, 14, 15], "b": [16, 17, 18]}))
    assert c._read(offset=5555) is None

    # iter
    for d, g in zip(c.iter(), generator_caller()):
        assert_frame_equal(d, g)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



"""Utilities for bivariate copulas."""

import numpy as np


def split_matrix(X):
    """Split an (n,2) numpy.array into two vectors.

    Args:
        X(numpy.array): Matrix of shape (n,2)

    Returns:
        tuple[numpy.array]: Both of shape (n,)

    """
    if len(X):
        return X[:, 0], X[:, 1]

    return np.array([]), np.array([])



from __future__ import annotations

import json
from collections import defaultdict
from collections.abc import Iterable
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, Literal, Set, Union

import pandas as pd
from pydantic import BaseModel, Field, field_validator

from sdgx.data_loader import DataLoader
from sdgx.data_models.inspectors.base import RelationshipInspector
from sdgx.data_models.inspectors.manager import InspectorManager
from sdgx.exceptions import MetadataInitError, MetadataInvalidError
from sdgx.models.components.utils import StrValuedBaseEnum
from sdgx.utils import logger


class CategoricalEncoderType(StrValuedBaseEnum):
    ONEHOT = "onehot"
    LABEL = "label"
    FREQUENCY = "frequency"


class Metadata(BaseModel):
    """
    Metadata is mainly used to describe the data types of all columns in a single data table.

    For each column, there should be an instance of the Data Type object.

    .. Note::

        Use ``get``, ``set``, ``add``, ``delete`` to update tags in the metadata. And use `query` for querying a column for its tags.

    Args:
        primary_keys(List[str]): The primary key, a field used to uniquely identify each row in the table.
        The primary key of each row must be unique and not empty.

        column_list(list[str]): list of the comlumn name in the table, other columns lists are used to store column information.
    """

    primary_keys: Set[str] = set()
    """
    primary_keys is used to store single primary key or composite primary key
    """

    column_list: List[str] = Field(default_factory=list, title="The List of Column Names")

    """"
    column_list is the actual value of self.column_list
    """

    @field_validator("column_list")
    @classmethod
    def check_column_list(cls, value) -> Any:
        # check if v has duplicate element
        if len(value) == len(set(value)):
            return value
        raise MetadataInitError("column_list has duplicate element!")

    column_inspect_level: Dict[str, int] = defaultdict(lambda: 10)
    """
    column_inspect_level is used to store every inspector's level, to specify the true type of each column.
    """

    pii_columns: Set[str] = set()
    """
    pii_columns is used to store all PII columns' name
    """

    # other columns lists are used to store column information
    # here are 6 basic data types
    id_columns: Set[str] = set()
    int_columns: Set[str] = set()
    float_columns: Set[str] = set()
    bool_columns: Set[str] = set()
    discrete_columns: Set[str] = set()
    datetime_columns: Set[str] = set()
    const_columns: Set[str] = set()
    datetime_format: Dict = defaultdict(str)
    numeric_format: Dict = defaultdict(list)

    # def columns encoder, even not matched categorical_threshold.
    categorical_encoder: Union[Dict[str, CategoricalEncoderType], None] = defaultdict(str)
    # if greater than categorical_threshold, encoder is the mapped.
    categorical_threshold: Union[Dict[int, CategoricalEncoderType], None] = None

    # version info
    version: str = "1.0"
    _extend: Dict[str, Set[str]] = defaultdict(set)
    """
    For extend information, use ``get`` and ``set``
    """

    def get_column_encoder_by_categorical_threshold(
        self, num_categories: int
    ) -> Union[CategoricalEncoderType, None]:
        encoder_type = None
        if self.categorical_threshold is None:
            return encoder_type

        for threshold in sorted(self.categorical_threshold.keys()):
            if num_categories > threshold:
                encoder_type = self.categorical_threshold[threshold]
            else:
                break
        return encoder_type

    def get_column_encoder_by_name(self, column_name) -> Union[CategoricalEncoderType, None]:
        encoder_type = None
        if self.categorical_encoder and column_name in self.categorical_encoder:
            encoder_type = self.categorical_encoder[column_name]
        return encoder_type

    @property
    def tag_fields(self) -> Iterable[str]:
        """
        Return all tag fields in this metadata.
        """

        return chain(
            (k for k in self.model_fields if k.endswith("_columns")),
            (k for k in self._extend.keys() if k.endswith("_columns")),
        )

    @property
    def format_fields(self) -> Iterable[str]:
        """
        Return all tag fields in this metadata.
        """

        return chain(
            (k for k in self.model_fields if k.endswith("_format")),
            (k for k in self._extend.keys() if k.endswith("_format")),
        )

    def __eq__(self, other):
        if not isinstance(other, Metadata):
            return super().__eq__(other)
        return (
            set(self.tag_fields) == set(other.tag_fields)
            and all(
                self.get(key) == other.get(key)
                for key in set(chain(self.tag_fields, other.tag_fields))
            )
            and all(
                self.get(key) == other.get(key)
                for key in set(chain(self.format_fields, other.format_fields))
            )
            and self.version == other.version
        )

    def query(self, field: str) -> Iterable[str]:
        """
        Query all tags of a field.

        Args:
            field(str): The field to query.

        Example:

            .. code-block:: python

                # Assume that user_id looks like 1,2,3,4
                m.query("user_id") == ["id_columns", "numeric_columns"]
        """
        return (k for k in self.tag_fields if field in self.get(k))

    def get(self, key: str) -> Set[str]:
        """
        Get all tags by key.

        Args:
            key(str): The key to get.

        Example:

            .. code-block:: python

                # Get all id columns
                m.get("id_columns") == {"user_id", "ticket_id"}
        """

        if key == "_extend":
            raise MetadataInitError("Cannot get _extend directly")

        return getattr(self, key) if key in self.model_fields else self._extend[key]

    def set(self, key: str, value: Any):
        """
        Set tags, will convert value to set if value is not a set.

        Args:
            key(str): The key to set.
            value(Any): The value to set.

        Example:

            .. code-block:: python

                # Set all id columns
                m.set("id_columns", {"user_id", "ticket_id"})
        """

        if key == "_extend":
            raise MetadataInitError("Cannot set _extend directly")

        old_value = self.get(key)
        if (
            key in self.model_fields
            and key not in self.tag_fields
            and key not in self.format_fields
        ):
            raise MetadataInitError(
                f"Set {key} not in tag_fields, try set it directly as m.{key} = value"
            )

        if isinstance(old_value, Iterable) and not isinstance(old_value, str):
            # list, set, tuple...
            value = value if isinstance(value, Iterable) and not isinstance(value, str) else [value]
            try:
                value = type(old_value)(value)
            except TypeError as e:
                if type(old_value) == defaultdict:
                    value = dict(value)
                else:
                    raise e

        if key in self.model_fields:
            setattr(self, key, value)
        else:
            self._extend[key] = value

    def add(self, key: str, values: str | Iterable[str]):
        """
        Add tags.

        Args:
            key(str): The key to add.
            values(str | Iterable[str]): The value to add.

        Example:

            .. code-block:: python

                # Add all id columns
                m.add("id_columns", "user_id")
                m.add("id_columns", "ticket_id")
                # OR
                m.add("id_columns", ["user_id", "ticket_id"])
                # OR
                # add datetime format
                m.add('datetime_format',{"col_1": "%Y-%m-%d %H:%M:%S", "col_2": "%d %b %Y"})
        """

        values = (
            values if isinstance(values, Iterable) and not isinstance(values, str) else [values]
        )

        # dict support,  this prevents the value in the key-value pair from being discarded
        if isinstance(values, dict):
            # already in fields that contains dict
            if key in list(self.format_fields):
                self.get(key).update(values)

            # in extend
            if self._extend.get(key, None) is None:
                self._extend[key] = values
            else:
                self._extend[key].update(values)
            return

        for value in values:
            self.get(key).add(value)

    def delete(self, key: str, value: str):
        """
        Delete tags.

        Args:
            key(str): The key to delete.
            value(str): The value to delete.

        Example:

            .. code-block:: python

                # Delete misidentification id columns
                m.delete("id_columns", "not_an_id_columns")

        """
        try:
            self.get(key).remove(value)
        except KeyError:
            pass

    def update(self, attributes: dict[str, Any]):
        """
        Update tags.
        """
        for k, v in attributes.items():
            self.add(k, v)

        return self

    @classmethod
    def from_dataloader(
        cls,
        dataloader: DataLoader,
        max_chunk: int = 10,
        primary_keys: Set[str] = None,
        include_inspectors: Iterable[str] | None = None,
        exclude_inspectors: Iterable[str] | None = None,
        inspector_init_kwargs: dict[str, Any] | None = None,
        check: bool = False,
    ) -> "Metadata":
        """Initialize a metadata from DataLoader and Inspectors

        Args:
            dataloader(DataLoader): the input DataLoader.
            max_chunk(int): max chunk count.
            primary_keys(list[str]): primary keys, see :class:`~sdgx.data_models.metadata.Metadata` for more details.
            include_inspectors(list[str]): data type inspectors used in this metadata (table).
            exclude_inspectors(list[str]): data type inspectors NOT used in this metadata (table).
            inspector_init_kwargs(dict): inspector args.
        """
        logger.info("Inspecting metadata...")
        im = InspectorManager()
        exclude_inspectors = exclude_inspectors or []
        exclude_inspectors.extend(
            name
            for name, inspector_type in im.registed_inspectors.items()
            if issubclass(inspector_type, RelationshipInspector)
        )

        inspectors = im.init_inspcetors(
            include_inspectors, exclude_inspectors, **(inspector_init_kwargs or {})
        )
        # set all inspectors not ready
        for inspector in inspectors:
            inspector.ready = False
        for i, chunk in enumerate(dataloader.iter()):
            for inspector in inspectors:
                if not inspector.ready:
                    inspector.fit(chunk)
            if all(i.ready for i in inspectors) or i > max_chunk:
                break

        if primary_keys is None:
            primary_keys = set()

        metadata = Metadata(primary_keys=primary_keys, column_list=dataloader.columns())
        for inspector in inspectors:
            inspect_res = inspector.inspect()
            # update column type
            metadata.update(inspect_res)
            # update pii column
            if inspector.pii:
                for each_key in inspect_res:
                    metadata.update({"pii_columns": inspect_res[each_key]})
            # update inspect level
            for each_key in inspect_res:
                if "columns" in each_key:
                    metadata.column_inspect_level[each_key] = inspector.inspect_level

        if not primary_keys:
            metadata.update_primary_key(metadata.id_columns)

        if check:
            metadata.check()
        return metadata

    @classmethod
    def from_dataframe(
        cls,
        df: pd.DataFrame,
        include_inspectors: list[str] | None = None,
        exclude_inspectors: list[str] | None = None,
        inspector_init_kwargs: dict[str, Any] | None = None,
        check: bool = False,
    ) -> "Metadata":
        """Initialize a metadata from DataFrame and Inspectors

        Args:
            df(pd.DataFrame): the input DataFrame.
            include_inspectors(list[str]): data type inspectors used in this metadata (table).
            exclude_inspectors(list[str]): data type inspectors NOT used in this metadata (table).
            inspector_init_kwargs(dict): inspector args.
        """

        im = InspectorManager()
        exclude_inspectors = exclude_inspectors or []
        exclude_inspectors.extend(
            name
            for name, inspector_type in im.registed_inspectors.items()
            if issubclass(inspector_type, RelationshipInspector)
        )

        inspectors = im.init_inspcetors(
            include_inspectors, exclude_inspectors, **(inspector_init_kwargs or {})
        )
        for inspector in inspectors:
            inspector.fit(df)

        metadata = Metadata(primary_keys=[df.columns[0]], column_list=df.columns)
        for inspector in inspectors:
            inspect_res = inspector.inspect()
            # update column type
            metadata.update(inspect_res)
            # update pii column
            if inspector.pii:
                for each_key in inspect_res:
                    metadata.update({"pii_columns": inspect_res[each_key]})
            # update inspect level
            for each_key in inspect_res:
                if "columns" in each_key:
                    metadata.column_inspect_level[each_key] = inspector.inspect_level

        if check:
            metadata.check()
        return metadata

    def _dump_json(self) -> str:
        return self.model_dump_json(indent=4)

    def save(self, path: str | Path):
        """
        Save metadata to json file.
        """

        with path.open("w") as f:
            f.write(self._dump_json())

    @classmethod
    def loads(cls, attributes):
        return Metadata(**attributes)

    @classmethod
    def load(cls, path: str | Path) -> "Metadata":
        """
        Load metadata from json file.
        """

        path = Path(path).expanduser().resolve()
        attributes = json.load(path.open("r"))
        version = attributes.get("version", None)
        if version:
            cls.upgrade(version, attributes)

        m = Metadata(**attributes)

        return m

    @classmethod
    def upgrade(cls, old_version: str, fields: dict[str, Any]) -> None:
        pass

    def check_single_primary_key(self, input_key: str):
        """Check whether a primary key in column_list and has ID data type.

        Args:
            input_key(str): the input primary_key str
        """

        if input_key not in self.column_list:
            raise MetadataInvalidError(f"Primary Key {input_key} not Exist in columns.")

    def get_all_data_type_columns(self):
        """Get all column names from `self.xxx_columns`.

        All Lists with the suffix _columns in model fields and extend fields need to be collected.
        All defined column names will be counted.

        Returns:
            all_dtype_cols(set): set of all column names.
        """
        all_dtype_cols = set()

        # search the model fields and extend fields
        for each_key in list(self.model_fields.keys()) + list(self._extend.keys()):
            if each_key.endswith("_columns"):
                column_names = self.get(each_key)
                all_dtype_cols = all_dtype_cols.union(set(column_names))
        return all_dtype_cols

    def check(self):
        """Checks column info.

        When passing as input to the next module, perform necessary checks, including:
            -Is the primary key correctly defined(in column list) and has ID data type.
            -Is there any missing definition of each column in table.
            -Are there any unknown columns that have been incorrectly updated.
        """
        # check primary key in column_list
        for each_key in self.primary_keys:
            self.check_single_primary_key(each_key)

        # for single primary key, it should has ID type
        if len(self.primary_keys) == 1 and list(self.primary_keys)[0] not in self.id_columns:
            raise MetadataInvalidError(f"Primary Key {self.primary_keys} should has ID DataType.")

        all_dtype_columns = self.get_all_data_type_columns()

        # check missing columns
        if set(self.column_list) - set(all_dtype_columns):
            raise MetadataInvalidError(
                f"Undefined data type for column {set(self.column_list) - set(all_dtype_columns)}."
            )

        # check unfamiliar columns in dtypes
        if set(all_dtype_columns) - set(self.column_list):
            raise MetadataInvalidError(
                f"Found undefined column: {set(all_dtype_columns) - set(self.column_list)}."
            )

        if self.categorical_encoder is not None:
            for i in self.categorical_encoder.keys():
                if not isinstance(i, str) or i not in self.discrete_columns:
                    raise MetadataInvalidError(
                        f"categorical_encoder key {i} is invalid, it should be an str and is a discrete column name."
                    )
            if self.categorical_encoder.values() not in CategoricalEncoderType:
                raise MetadataInvalidError(
                    f"In categorical_encoder values, categorical encoder type invalid, now supports {list(CategoricalEncoderType)}."
                )

        if self.categorical_threshold is not None:
            for i in self.categorical_threshold.keys():
                if not isinstance(i, int) or i < 0:
                    raise MetadataInvalidError(
                        f"categorical threshold {i} is invalid, it should be an positive int."
                    )
            if self.categorical_threshold.values() not in CategoricalEncoderType:
                raise MetadataInvalidError(
                    f"In categorical_threshold values, categorical encoder type invalid, now supports {list(CategoricalEncoderType)}."
                )

        logger.debug("Metadata check succeed.")

    def update_primary_key(self, primary_keys: Iterable[str] | str):
        """Update the primary key of the table

        When update the primary key, the original primary key will be erased.

        Args:
            primary_keys(Iterable[str]): the primary keys of this table.
        """

        if not isinstance(primary_keys, Iterable) and not isinstance(primary_keys, str):
            raise MetadataInvalidError("Primary key should be Iterable or str.")
        primary_keys = set(primary_keys if isinstance(primary_keys, Iterable) else [primary_keys])

        if not primary_keys.issubset(set(self.column_list)):
            raise MetadataInvalidError("Primary key not exist in table columns.")

        self.primary_keys = primary_keys

        logger.info(f"Primary Key updated: {primary_keys}.")

    def dump(self):
        """Dump model dict, can be used in downstream process, like processor.

        Returns:
            dict: dumped dict.
        """
        model_dict = self.model_dump()
        model_dict["column_data_type"] = {}
        for each_col in self.column_list:
            model_dict["column_data_type"][each_col] = self.get_column_data_type(each_col)
        return model_dict

    def get_column_data_type(self, column_name: str):
        """Get the exact type of specific column.
        Args:
            column_name(str): The query colmun name.
        Returns:
            str: The data type query result.
        """
        if column_name not in self.column_list:
            raise MetadataInvalidError(f"Column {column_name}not exists in metadata.")
        current_type = None
        current_level = 0
        # find the dtype who has most high inspector level
        for each_key in list(self.model_fields.keys()) + list(self._extend.keys()):
            if (
                each_key != "pii_columns"
                and each_key.endswith("_columns")
                and column_name in self.get(each_key)
                and current_level < self.column_inspect_level[each_key]
            ):
                current_level = self.column_inspect_level[each_key]
                current_type = each_key
        if not current_type:
            raise MetadataInvalidError(f"Column {column_name} has no data type.")
        return current_type.split("_columns")[0]

    def get_column_pii(self, column_name: str):
        """Return if a column is a PII column.
        Args:
            column_name(str): The query colmun name.
        Returns:
            bool: The PII query result.
        """
        if column_name not in self.column_list:
            raise MetadataInvalidError(f"Column {column_name}not exists in metadata.")
        if column_name in self.pii_columns:
            return True
        return False

    def change_column_type(
        self, column_names: str | List[str], column_original_type: str, column_new_type: str
    ):
        """Change the type of column."""
        if not column_names:
            return
        if isinstance(column_names, str):
            column_names = [column_names]
        all_fields = list(self.tag_fields)
        original_type = f"{column_original_type}_columns"
        new_type = f"{column_new_type}_columns"
        if original_type not in all_fields:
            raise MetadataInvalidError(f"Column type {column_original_type} not exist in metadata.")
        if new_type not in all_fields:
            raise MetadataInvalidError(f"Column type {column_new_type} not exist in metadata.")
        type_columns = self.get(original_type)
        diff = set(column_names).difference(type_columns)
        if diff:
            raise MetadataInvalidError(f"Columns {column_names} not exist in {original_type}.")
        self.add(new_type, column_names)
        type_columns = type_columns.difference(column_names)
        self.set(original_type, type_columns)

    def remove_column(self, column_names: List[str] | str):
        """
        Remove a column from all columns type.
        Args:
            column_names: List[str]: To removed columns name list.
        """
        if not column_names:
            return
        if isinstance(column_names, str):
            column_names = [column_names]
        column_names = frozenset(column_names)
        inter = column_names.intersection(self.column_list)
        if not inter:
            raise MetadataInvalidError(f"Columns {inter} not exist in metadata.")

        def do_remove_columns(key, get=True, to_removes=column_names):
            obj = self
            if get:
                target = obj.get(key)
            else:
                target = getattr(obj, key)
            res = None
            if isinstance(target, list):
                res = [item for item in target if item not in to_removes]
            elif isinstance(target, dict):
                if key == "numeric_format":
                    obj.set(
                        key,
                        {k: {v2 for v2 in v if v2 not in to_removes} for k, v in target.items()},
                    )
                else:
                    res = {k: v for k, v in target.items() if k not in to_removes}
            elif isinstance(target, set):
                res = target.difference(to_removes)

            if res is not None:
                if get:
                    obj.set(key, res)
                else:
                    setattr(obj, key, res)

        to_remove_attribute = list(self.tag_fields)
        to_remove_attribute.extend(list(self.format_fields))
        for attr in to_remove_attribute:
            do_remove_columns(attr)
        for attr in ["column_list", "primary_keys"]:
            do_remove_columns(attr, False)
        self.check()



from __future__ import annotations

from functools import cached_property
from typing import Any, Generator

import pandas as pd

from sdgx.cachers.base import Cacher, NoCache
from sdgx.cachers.disk_cache import DiskCache
from sdgx.cachers.manager import CacherManager
from sdgx.data_connectors.base import DataConnector
from sdgx.data_connectors.dataframe_connector import DataFrameConnector
from sdgx.data_connectors.generator_connector import GeneratorConnector
from sdgx.exceptions import DataLoaderInitError
from sdgx.utils import cache


class DataLoader:
    """
    Combine :ref:`Cacher` and :ref:`DataConnector` to load data in an efficient way.

    Default Cacher is :ref:`DiskCache`. Use ``cacher`` or ``cache_mode`` to specify a :ref:`Cacher`.

    GeneratorConnector must combine with Cacher, we will warmup cache for generator to support random access.

    Args:
        data_connector (:ref:`DataConnector`): The data connector
        chunksize (int, optional): The chunksize of the cacher. Defaults to 1000.
        cacher (:ref:`Cacher`, optional): The cacher. Defaults to None.
        cache_mode (str, optional): The cache mode(cachers' name). Defaults to "DiskCache", more info in :ref:`DiskCache`.
        cacher_kwargs (dict, optional): The kwargs for cacher. Defaults to None
        identity (str, optional): The identity of the data source.
            When using :ref:`GeneratorConnector`, it can be pointed to the original data source, makes it possible to work with :ref:`MetadataCombiner`.

    Example:

        Load and cache data from existing csv file or other data source.

        .. code-block:: python

            from sdgx.data_loader import DataLoader
            from sdgx.data_connectors.csv_connector import CsvConnector
            from sdgx.utils import download_demo_data

            dataset_csv = download_demo_data()
            data_connector = CsvConnector(path=dataset_csv)

            # Use DataConnector to initialize

            dataloader = DataLoader(data_connector)

            # Access data

            dataloader.load_all()  # This will read all data from csv, and cache it.
            dataloader.load_all()  # This will read all data from cache.

            dataloader[:10] # dataloader support slicing

            for df in dataloader.iter():  # dataloader support iteration
                print(df.shape)

    Advanced usage:

        Load and cache data from a generator.

        .. code-block:: python

            from sdgx.data_loader import DataLoader
            from sdgx.data_connectors.generator_connector import GeneratorConnector

            def generator() -> Generator[pd.DataFrame, None, None]:
                for i in range(100):
                    yield pd.DataFrame({"a": [i], "b": [i]})

            data_connector = GeneratorConnector(generator)

            # Use DataConnector to initialize.
            # Generator is not support random access, but we can achieve it by caching.
            dataloader = DataLoader(data_connector)

            # Access data
            dataloader.load_all()  # This will read all data from cache
            dataloader.load_all()  # This will read all data from cache.

            dataloader[:10] # dataloader support slicing

            for df in dataloader.iter():  # dataloader support iteration
                print(df.shape)


    """

    DEFAULT_CACHER_INITIAL = DiskCache

    def __init__(
        self,
        data_connector: DataConnector,
        chunksize: int = 10000,
        cacher: Cacher | str | type[Cacher] | None = None,
        cacher_kwargs: None | dict[str, Any] = None,
        identity: str | None = None,
    ) -> None:
        if isinstance(data_connector, DataFrameConnector):
            self.DEFAULT_CACHER = NoCache
        else:
            self.DEFAULT_CACHER = DataLoader.DEFAULT_CACHER_INITIAL

        self.data_connector = data_connector
        self.chunksize = chunksize
        self.cache_manager = CacherManager()
        self.identity = identity or self.data_connector.identity or str(id(self))

        if not cacher_kwargs:
            cacher_kwargs = {}
        cacher_kwargs.setdefault("blocksize", self.chunksize)
        cacher_kwargs.setdefault("identity", self.data_connector.identity)
        if isinstance(cacher, Cacher):
            self.cacher = cacher
        elif isinstance(cacher, str) or isinstance(cacher, type):
            self.cacher = self.cache_manager.init_cacher(cacher, **cacher_kwargs)
        else:
            self.cacher = self.cache_manager.init_cacher(self.DEFAULT_CACHER, **cacher_kwargs)

        self.cacher.clear_invalid_cache()

        if isinstance(data_connector, GeneratorConnector):
            if isinstance(self.cacher, NoCache):
                raise DataLoaderInitError("NoCache can't be used with GeneratorConnector")
            # Warmup cache for generator, this allows random access
            self.load_all()

    def iter(self) -> Generator[pd.DataFrame, None, None]:
        """
        Load data from cache in chunk.
        """
        for d in self.cacher.iter(self.chunksize, self.data_connector):
            yield d

    def keys(self) -> list:
        """
        Same as ``columns``
        """
        return self.data_connector.keys()

    def columns(self) -> list:
        """
        Peak columns.

        Returns:
            list: name of columns
        """
        return self.data_connector.columns()

    def load_all(self) -> pd.DataFrame:
        """
        Load all data from cache.
        """
        return self.cacher.load_all(self.data_connector)

    def finalize(self, clear_cache=False) -> None:
        """
        Finalize the dataloader.
        """
        self.data_connector.finalize()
        if clear_cache:
            self.cacher.clear_cache()

    def __getitem__(self, key: int | slice | list) -> pd.DataFrame:
        """
        Support get data by index and slice.
        """
        if isinstance(key, int):
            return self.cacher.load(
                offset=(key // self.chunksize) * self.chunksize,
                chunksize=self.chunksize,
                data_connector=self.data_connector,
            )[0]

        if isinstance(key, list):
            return pd.concat((d[key] for d in self.iter()), ignore_index=True)

        assert isinstance(key, slice)
        start = key.start or 0
        stop = key.stop or len(self)
        step = key.step or 1

        offset = (start // self.chunksize) * self.chunksize
        n_iter = ((stop - start) // self.chunksize) + 1

        tables = (
            self.cacher.load(
                offset=offset + i * self.chunksize,
                chunksize=self.chunksize,
                data_connector=self.data_connector,
            )
            for i in range(n_iter)
        )

        return pd.concat(tables, ignore_index=True)[start - offset : stop - offset : step]

    @cache
    def __len__(self):
        return sum(len(l) for l in self.iter())

    @cached_property
    def shape(self):
        return (len(self), len(self.columns()))



"""RDT Exceptions."""


class NotFittedError(Exception):
    """Error to raise when ``transform`` or ``reverse_transform`` are used before fitting."""


class Error(Exception):
    """Error to raise when ``HyperTransformer`` produces a controlled error message."""



"""GaussianUnivariate module."""

import numpy as np
from scipy.stats import norm

from sdgx.models.components.sdv_copulas.univariate.base import (
    BoundedType,
    ParametricType,
    ScipyModel,
)


class GaussianUnivariate(ScipyModel):
    """Gaussian univariate model."""

    PARAMETRIC = ParametricType.PARAMETRIC
    BOUNDED = BoundedType.UNBOUNDED

    MODEL_CLASS = norm

    def _fit_constant(self, X):
        self._params = {"loc": np.unique(X)[0], "scale": 0}

    def _fit(self, X):
        self._params = {"loc": np.mean(X), "scale": np.std(X)}

    def _is_constant(self):
        return self._params["scale"] == 0

    def _extract_constant(self):
        return self._params["loc"]



from __future__ import annotations

from sdgx.data_processors.base import DataProcessor


class Filter(DataProcessor):
    """
    Base class for all data filters.

    Filter is a module used to apply rules and remove sampled data that does not conform to the rules.
    """
